import pandas as pd
import scanpy as sc
import muon as mu
import argparse
# import h5py
import os

from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice

import multiprocessing 
threads_available = multiprocessing.cpu_count()

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_anndata',
                    default='adata_log1p.h5ad',
                    help='')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_combat.csv',
                    help='')
parser.add_argument('--integration_col', default='batch', help='')
parser.add_argument('--n_threads', default=1,
                    help="num threads to use for neighbor computations")
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs")
parser.add_argument('--neighbors_method',
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k',
                    help="neighbors k")
parser.add_argument('--neighbors_metric',
                    help="neighbor metric, e.g. euclidean or cosine")
parser.add_argument('--modality',
                    help="")


args, opt = parser.parse_known_args()



L.info("reading data and starting integration pipeline with script: ")
L.info(os.path.basename(__file__))
L.info("Running with options: %s", args)
L.info("threads available: %s", threads_available)
# this should be an object that contains the full log normalised data (adata_log1p.h5ad)
# prior to hvgs and filtering
#adata = read_anndata(args.input_anndata, use_muon=use_muon, modality="rna")
mdata = mu.read(args.input_anndata)
adata = mdata.mod[args.modality] 

# combat can't integrate on 2+ variables, so create a fake column with combined information
columns = [x.strip() for x in args.integration_col.split(",")]
if len(columns) > 1: 
    L.info("using 2 columns to integrate on more variables")
    #comb_columns = "_".join(columns)
    adata.obs["comb_columns"] = adata.obs[columns].apply(lambda x: '|'.join(x), axis=1)

    # make sure that batch is a categorical
    adata.obs["comb_columns"] = adata.obs["comb_columns"].astype("category")
    # run combat
    sc.pp.combat(adata, key="comb_columns")
    
else:

    # run combat
    sc.pp.combat(adata, key=args.integration_col)

L.info("integration run now calculate umap")
sc.pp.pca(adata, use_highly_variable=True, svd_solver='arpack') #use same args as plot_pca.py

# run neighbours and umap 
run_neighbors_method_choice(adata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=int(args.neighbors_n_pcs), 
    metric=args.neighbors_metric, 
    use_rep='X_pca',
    nthreads=max([threads_available, 6]))


sc.tl.umap(adata)
L.info("done umap, saving stuff")
#write out
umap = pd.DataFrame(adata.obsm['X_umap'], adata.obs.index)
umap.to_csv(args.output_csv)

adata.write("tmp/combat_scaled_adata_" + args.modality + ".h5ad")


L.info("Done")




import argparse
import pandas as pd
import re


from panpipes.funcs.processing import  extract_parameter_from_fname

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("test logging message")

# parse arguments
parser = argparse.ArgumentParser()
parser.add_argument('--input_files_str',
                    default='',
                    help='')
parser.add_argument('--output_file',
                    default='',
                    help='')
parser.add_argument('--sample_prefix',
                    default='',
                    help='')
parser.add_argument('--clusters_or_markers',
                    default='',
                    help='')
parser.set_defaults(verbose=True)
args = parser.parse_args()

L.info(args)

infiles = re.split(',', args.input_files_str)
if args.clusters_or_markers == "clusters":
    combined_csv = pd.concat([pd.read_csv(f, sep='\t', index_col=0) for f in infiles], axis=1)
    # get colnames
    cnames = []
    for f in infiles:
        alg = extract_parameter_from_fname(f, 'alg', prefix=args.sample_prefix)
        res = extract_parameter_from_fname(f, 'res', prefix=args.sample_prefix)
        cnames.append(alg + '_res_' + str(res))
    combined_csv.to_csv(args.output_file, sep='\t', header=cnames, index=True)


if args.clusters_or_markers == "markers":
    li = []
    all_markers_file = re.sub("_top", "_all", args.output_file)
    excel_file = re.sub("_top.txt.gz", "_all.xlsx", args.output_file)
    excel_file_top = re.sub("_top.txt.gz", "_top.xlsx", args.output_file)
    with pd.ExcelWriter(excel_file) as writer:
        with pd.ExcelWriter(excel_file_top) as writer2:
            for ff in infiles:
                # print(os.path.join(in_path, ff))
                df = pd.read_csv(ff, sep='\t')
                # add a cluster column
                clust_val = extract_parameter_from_fname(ff, "cluster", prefix=args.sample_prefix)
                sname = "cluster" + str(clust_val)
                df['cluster'] = clust_val
                li.append(df)
                df.to_excel(writer, sheet_name=sname, index=False)
                df_sub = df[df['p.adj.bonferroni'] < 0.05]
                df_sub = df_sub[df_sub['avg_logFC'] > 0]
                df_sub.to_excel(writer2, sheet_name=sname, index=False)
    frame = pd.concat(li, axis=0, ignore_index=True)
    frame.to_csv(all_markers_file, sep='\t', header=True, index=False)
    frame_sub = frame[frame['p.adj.bonferroni'] < 0.05]
    frame_sub = frame_sub[frame_sub['avg_logFC'] > 0]
    frame_sub.to_csv(args.output_file, sep='\t', header=True, index=False)


L.info("Done")




##
# Scanorama is designed to be used in scRNA-seq pipelines downstream of noise-reduction methods,
# including those for imputation and highly-variable gene filtering.
##

import pandas as pd
import scanpy as sc
import os
# import scanorama
import scanpy.external as sce
import argparse
import muon as mu

from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# check for threads number
import multiprocessing 
threads_available = multiprocessing.cpu_count()

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_scanorama.csv',
                    help='')
parser.add_argument('--integration_col', default='batch',
                    help='')
parser.add_argument('--n_threads', default=1,
                    help="num threads to use for neighbor computations")
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs")
parser.add_argument('--neighbors_method',
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k',
                    help="neighbors k")
parser.add_argument('--neighbors_metric',
                    help="neighbor metric, e.g. euclidean or cosine")
parser.add_argument('--batch_size',
                    help="batch size scanorama parameter, default 5000", default="5000")
parser.add_argument('--modality',
                    help="modality")



args, opt = parser.parse_known_args()



L.info("Running with options: %s", args)

# Scanorama is designed to be used in scRNA-seq pipelines downstream of noise-reduction methods,
# including those for imputation and highly-variable gene filtering.
L.info("reading data and starting integration pipeline with script: ")
L.info(os.path.basename(__file__))

#adata = read_anndata(args.input_anndata, use_muon=use_muon, modality="rna")
mdata = mu.read(args.input_anndata)
adata = mdata.mod[args.modality] 
bcs = adata.obs_names.tolist()

# scanorama can't integrate on 2+ variables, so create a fake column with combined information
columns = [x.strip() for x in args.integration_col.split(",")]
if len(columns) > 1:
    L.info("using 2 columns to integrate on more variables")
    # comb_columns = "_".join(columns)
    adata.obs["comb_columns"] = adata.obs[columns].apply(lambda x: '|'.join(x), axis=1)

    # make sure that batch is a categorical
    adata.obs["comb_columns"] = adata.obs["comb_columns"].astype("category")
    adata.obs['batch'] = adata.obs["comb_columns"]
else:
    if args.integration_col != "batch":
        adata.obs['batch'] = adata.obs[args.integration_col]

batches = adata.obs['batch'].unique()
L.info("define scanorama inputs based on batches and number of PCs")
# need contiguous batches
scanorama_input = adata[adata.obs.sort_values(by="batch").index.tolist(), :]

# filter by HVGs to make it equivalent to the old scripts,
# which inputted the scaled object after filtering by hvgs.
scanorama_input = scanorama_input[:, scanorama_input.var.highly_variable]
#also filter the X_PCA to be the number of PCs we actually want to use
scanorama_input.obsm['X_pca'] = scanorama_input.obsm['X_pca'][:,0:int(args.neighbors_n_pcs)]


# run scanoramam using the scanpy integrated approach
L.info("run_scanorama")

sce.pp.scanorama_integrate(scanorama_input, key='batch', batch_size=int(args.batch_size))
L.info("scanorama done")

# not integrated

# old method (simplified)
# alldata = {}
# for bb in batches:
#     alldata[bb] = scanorama_input[scanorama_input.obs['batch'] == bb]
#
# adatas = list(alldata.values())
# X_scanorama = scanorama.integrate_scanpy(adatas)
# scanorama_input.obsm['X_scanorama'] = np.concatenate(X_scanorama)

# put into the original order
scanorama_input = scanorama_input[bcs, :]
# check it worked
if all(scanorama_input.obs_names == bcs):
    L.info("barcode order is correct")
else:
    L.debug("barcode order in  batch corrected object is incorrect")
    sys.exit("barcode order in  batch corrected object is incorrect")

L.info(adata)
# add to the AnnData object
adata.obsm["X_scanorama"] = scanorama_input.obsm["X_scanorama"]

L.info("integration run now calculate umap")

if int(args.neighbors_n_pcs) > adata.obsm['X_scanorama'].shape[1]:
    L.warn(f"N PCs is larger than X_scanorama dimensions, reducing n PCs to  {adata.obsm['X_scanorama'].shape[1] -1}")
n_pcs= min(int(args.neighbors_n_pcs), adata.obsm['X_scanorama'].shape[1]-1)

# run neighbours and umap 
run_neighbors_method_choice(adata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=n_pcs, 
    metric=args.neighbors_metric, 
    use_rep='X_scanorama',
    nthreads=max([threads_available, 6]))


sc.tl.umap(adata)
L.info("done umap, saving stuff")
# write out
umap = pd.DataFrame(adata.obsm['X_umap'], adata.obs.index)
umap.to_csv(args.output_csv)

# save the scanorama dim reduction in case scanorama is our favourite
adata.write("tmp/scanorama_scaled_adata.h5ad")

L.info("done")




from random import sample
import warnings
import re
import os
import pandas as pd
import sys
import h5py
from typing import Optional, Literal
from scanpy import read_10x_mtx, read_10x_h5, read_h5ad, read_text, read_csv, read_hdf

import muon as mu
import logging
import scirpy as ir 
from itertools import chain
from .processing import concat_adata_list, intersection
from anndata import AnnData
import numpy as np
from os import PathLike

def update_cellranger_col(path,  raw=False, method="count", sample_id=""):
    if method =="count":
        # this assumes that we get a cellranger count base path and we want file path for sc.read_10x_h5 or similar
        if raw:
            append_str = "raw_feature_bc_matrix"
        else:
            append_str = "filtered_feature_bc_matrix"
    elif method == "multi":
        if raw:
            append_str = os.path.join("multi", "count", "raw_feature_bc_matrix")
        else:
            append_str = os.path.join("per_sample_outs", sample_id, "count", "sample_filtered_feature_bc_matrix")
    else:
        sys.exit("specify cellranger filetype as cellranger_count or cellranger_multi")
    # is there a h5 file available
    if os.path.exists(os.path.join(path, append_str)) == False:
        sys.exit("cellranger path not found: %s" % os.path.join(path, append_str))
    if os.path.exists(os.path.join(path, append_str + '.h5')):
        append_str = append_str + ".h5"
        filetype = "10X_h5"
    else:
        filetype = "cellranger"
    path = os.path.join(path, append_str)
    return path, filetype


def gen_load_anndata_jobs(caf, load_raw=False, mode_dictionary = {}, load_prot_from_raw=False):
    """
    Generate a load_adatas job for each line in submission.txt
    """
    for nn in range(0, caf.shape[0]):
        if pd.isna(caf['rna_path'][nn]):
                rna_path= None
                rna_filetype=None
        elif caf['rna_filetype'][nn]=="cellranger" and mode_dictionary["rna"]:
            rna_path, rna_filetype = update_cellranger_col(caf['rna_path'][nn], raw=load_raw, method="count")
        elif caf['rna_filetype'][nn]=="cellranger_multi" and mode_dictionary["rna"]:
            rna_path, rna_filetype = update_cellranger_col(caf['rna_path'][nn], raw=load_raw, method="multi", 
                                                            sample_id=caf['sample_id'][nn])
        else:
            rna_path, rna_filetype = caf[['rna_path', "rna_filetype"]].iloc[nn]
            if load_raw:
                rna_path = re.sub("filtered", "raw", rna_path)
        # manage the prot paths
        if ('prot_path' in caf.columns and mode_dictionary["prot"]):
            # check if its the same as the rna path (data in the same file)
            if pd.isna(caf['prot_path'][nn]):
                prot_path= None
                prot_filetype=None
            elif caf['prot_path'][nn] == caf['rna_path'][nn]:
                prot_path, prot_filetype = rna_path, rna_filetype
            elif caf['prot_filetype'][nn]=="cellranger":
                # we might want to load the raw here because we want to then subset by good gex (rna) barcodes, 
                # this is why the load_prot_from_raw argument exists
                prot_path, prot_filetype = update_cellranger_col(caf['prot_path'][nn], raw=load_prot_from_raw)
            elif caf['prot_filetype'][nn]=="cellranger_multi":
                # celranger multi has the same prot and gex (rna) barcodes
                prot_path, prot_filetype = update_cellranger_col(caf['prot_path'][nn], raw=load_raw, method="multi", 
                                                                sample_id=caf['sample_id'][nn])
            else:
                prot_path, prot_filetype = caf[['prot_path', "prot_filetype"]].iloc[nn]
                if load_prot_from_raw or load_raw:
                    prot_path = re.sub("filtered", "raw", prot_path)
        else:
            prot_path= None
            prot_filetype=None
        # load tcr_path
        if 'tcr_path' in caf.columns and mode_dictionary["tcr"] and pd.notna(caf['tcr_path'][nn]):
            tcr_path = caf['tcr_path'][nn]
            tcr_filetype = caf['tcr_filetype'][nn]
        else:
            tcr_path= None
            tcr_filetype=None
        if 'bcr_path' in caf.columns and mode_dictionary["bcr"] and pd.notna(caf['bcr_path'][nn]):
            bcr_path = caf['bcr_path'][nn]
            bcr_filetype = caf['bcr_filetype'][nn]
        else:
            bcr_path= None
            bcr_filetype=None
        if ('atac_path' in caf.columns and mode_dictionary["atac"]):
            if caf.shape[0] > 1:
                sys.exit("You can only submit one atac/multiome file at a time. To aggregate, see cellranger aggr.")
            if caf['atac_filetype'][nn]=="cellranger" :
                atac_path, atac_filetype = update_cellranger_col(caf['atac_path'][nn], raw=load_raw, method="count")
            else:
                atac_path = caf['atac_path'][nn]
                atac_filetype = caf['atac_filetype'][nn]
            if 'fragments_file' in caf.columns and pd.notna(caf['fragments_file'][nn]):
                fragments_file = caf['fragments_file'][nn]
            else:
                fragments_file = None
            if 'per_barcode_metrics_file' in caf.columns and pd.notna(caf['per_barcode_metrics_file'][nn]):
                per_barcode_metrics_file = caf['per_barcode_metrics_file'][nn]
            else:
                per_barcode_metrics_file = None
            if 'peak_annotation_file' in caf.columns and pd.notna(caf['peak_annotation_file'][nn]):    
                peak_annotation_file = caf['peak_annotation_file'][nn]
            else:
                peak_annotation_file = None
        else:
            atac_path= None
            atac_filetype=None
            fragments_file = None
            peak_annotation_file = None
            per_barcode_metrics_file = None
            
        if 'barcode_mtd_path' in caf.columns:
            cell_mtd_path = caf['barcode_mtd_path'][nn]
        else:
            cell_mtd_path = None
        # create the output file 
        outfile = "./tmp/" + caf['sample_id'][nn]
        if load_raw:
            outfile = outfile + "_raw.h5mu"
        else:
            outfile = outfile + ".h5mu"
        sample_id = caf['sample_id'][nn]
        yield rna_path, outfile, \
              sample_id, \
              rna_filetype,  \
              prot_path, prot_filetype, \
              tcr_path, tcr_filetype,  \
              bcr_path, bcr_filetype, \
              atac_path, atac_filetype, \
              fragments_file, per_barcode_metrics_file, peak_annotation_file, \
              cell_mtd_path
        

def gen_load_spatial_jobs(caf, mode_dictionary = {}, load_raw=True):
    """
    Generate a load_spatial job for each line in submission.txt
    """
    for nn in range(0, caf.shape[0]):
        if "spatial_path" in caf.columns and mode_dictionary["spatial"]:
            if pd.isna(caf["spatial_path"][nn]):
                spatial_path= None
                spatial_filetype = None
            else:
                spatial_path = caf["spatial_path"][nn]
            if caf['spatial_filetype'][nn]=="vizgen":
                spatial_filetype = caf['spatial_filetype'][nn]
                #path, counts and metadata are mandatory
                if pd.notna(caf["spatial_counts"][nn]):
                    spatial_counts= caf["spatial_counts"][nn]
                else:
                    spatial_counts = None
                if pd.notna(caf["spatial_metadata"][nn]):
                    spatial_metadata = caf["spatial_metadata"][nn]
                else: 
                    spatial_metadata = None
                #transformation is optional
                if pd.notna(caf["spatial_transformation"][nn]):
                    spatial_transformation = caf["spatial_transformation"][nn]
                else:
                    spatial_transformation = None
            elif caf['spatial_filetype'][nn]=="visium":
                spatial_metadata= None
                spatial_transformation = None
                spatial_filetype = caf['spatial_filetype'][nn]
                if pd.notna(caf["spatial_counts"][nn]):
                    spatial_counts= caf["spatial_counts"][nn]
                else:
                    spatial_counts = None  
        else:
            spatial_path= None
            spatial_filetype = None
            spatial_counts = None
            spatial_metadata = None
            spatial_transformation = None
            
        if 'barcode_mtd_path' in caf.columns:
            cell_mtd_path = caf['barcode_mtd_path'][nn] #not yielding this right now!
        else:
            cell_mtd_path = None
        # create the output file 
        outfile = "./tmp/" + caf['sample_id'][nn]
        if load_raw:
            outfile = outfile + "_raw.h5mu" 
        else:
            outfile = outfile + ".h5mu"
        sample_id = caf['sample_id'][nn]
        yield spatial_path,  outfile, \
              sample_id, spatial_filetype, spatial_counts, spatial_metadata, spatial_transformation


def read_anndata(
    fname: Optional[str] = None,
    use_muon: Optional[bool] = False, 
    modality: Literal['all', 'rna', 'prot','atac','rep'] = 'all'):
    logging.info("reading %s" % fname)
    # check fname file exists
    try:
        os.path.isfile(fname)
    except FileNotFoundError:
        sys.exit("anndata file not found")
    if use_muon is False:
        return read_h5ad(fname)
    else:
        logging.info("reading %s modality" % modality)
        if modality=="all":
            # note this just loads the rna part of the muon object
            return mu.read(fname)
        if modality=="rna":
            # note this just loads the rna part of the muon object
            return mu.read(fname + "/rna")
        elif modality=="prot":
            return mu.read(fname + "/prot")
        elif modality=="atac":
            return mu.read(fname + "/atac")
        elif modality=="rep":
            return mu.read(fname + "/rep")
        
        else:
            sys.exit("modality not found, must be one of 'all', 'rna', 'prot','atac', 'rep' ")


def write_anndata(adata,
                    fname, 
                    use_muon=False, 
                    modality: Literal['all', 'rna', 'prot'] = 'all'):
    if use_muon is False:
        adata.write(fname)
    else:
        if modality=="rna":
            # note this just saves the rna part of the muon object
            return mu.write(fname + "/rna", adata)
        if modality=="prot":
            # note this just save the prot part of the muon object
            return mu.write(fname + "/prot", adata)
        elif modality=="all":
            return mu.write(fname, adata)
        else:
            sys.exit("modality not found, must be one of 'all', 'rna', 'prot'")

# --- loading raw dat

def write_obs(mdata, output_prefix="./", output_suffix="_filtered_cell_metadata.tsv"):
        metafile = mdata.obs
        metafile["cellbarcode"] = mdata.obs.index
        savename= output_prefix + output_suffix
        metafile.to_csv(savename, sep='\t', index=True)

def check_submission_file(caf):
    if "filetype" in caf.columns or "path" in caf.columns:
        raise ValueError("you appear to be using the old notation for the sample submission file, \
        please update to use rna_path instead of path and rna_filetype instead of filetype")
    # check for required cols
    req_cols = ['sample_id', 'rna_path', 'rna_filetype']
    for rc in req_cols:
        if rc not in caf.columns:
            raise ValueError("required column %s missing from submission file" % rc)



def check_filetype(path, filetype):
    logging.debug(path)
    logging.debug(filetype)
    if filetype == "cellranger" :
        if os.path.exists(path + "/matrix.mtx.gz"):
            logging.info("cellranger matrix found")
        elif os.path.exists(path + "/matrix.mtx"):
            logging.info("cellranger matrix found")
        else:
            sys.exit("cellranger matrix file not found, have you specified the path correctly")
    else:
        ftype_checks = {
            "h5ad":r".h5ad$",
            "csv_matrix": ".csv",
            "txt_matrix" : ".txt",
            "10X_h5": ".h5",
            "hdf": ".h5",
            "cellranger_vdj": r".json|.csv",
            "vizgen": r".txt|.csv|.tsv" #suboptimal for now but roll with it
        }
        if filetype not in ftype_checks.keys():
            sys.exit("unknown filetype %s, please specify one of: %s" % filetype, ftype_checks.keys().join(", "))
        elif re.search(ftype_checks[filetype], path) is None:
            sys.exit("filetype does not match file suffix, check it is specified correctly \
                    (with %s as suffix (gzipped files ok for csv and tsv))" % ftype_checks[filetype])
        elif os.path.exists(path)==False:
            sys.exit("file %s does not exist" % path)
        else:
            logging.info("%s file exists" % path)


def read_scirpy(
    fname, 
    filetype: Literal['cellranger_vdj', 'tracer', 'bracer', 'airr' ] = 'cellranger_vdj'):
    """
    wrapper around all the scirpy io functions
    """
    # check fname file exists
    logging.info("loading repertoire using scirpy, filetype %s" % filetype)
    try:
        os.path.isfile(fname)
    except FileNotFoundError:
        sys.exit("vdj file not found: %s" % fname)
    if filetype ==  "cellranger_vdj":
        vdj = ir.io.read_10x_vdj(fname)
    elif filetype == "tracer":
        vdj = ir.io.read_tracer(fname)
    elif filetype ==  "bracer":
        vdj = ir.io.read_bracer(fname)
    elif filetype ==  "airr":
        vdj = ir.io.read_airr(fname)
    else:
        sys.exit("fileype not valid, must be one of 'cellranger_vdj', 'tracer', 'bracer', 'airr'")
    return vdj


def scp_read_10x_h5(filename: PathLike, library_keep=None, extended= False, *args, **kwargs) -> AnnData:
    """
    expanded sc.read_10x_h5 to filter for the library 
    adapted from https://github.com/scverse/muon/blob/master/muon/_prot/io.py
    """
    adata = read_10x_h5(filename, gex_only=False, *args, **kwargs)
    logging.debug(adata)
    logging.debug("filtering cellranger outputs to %s" % library_keep)
    if extended is not None and extended is True:
        import shutil
        logging.debug("copying %s" % filename )
        h5file = h5py.File(filename, "r")
        logging.debug("reading intervals")
        if "interval" in h5file["matrix"]["features"]:
            intervals = np.array(h5file["matrix"]["features"]["interval"]).astype(str)
            logging.debug("intervals %s" %intervals)
            adata.var["interval"] = intervals
            logging.debug(f"Added `interval` annotation for features from {filename}")
            logging.debug("updated adata %s" % adata)
            logging.debug("adata.var is %s" % adata.var)
            h5file.close()
        else:
            # Make sure the file is closed
            h5file.close()
    else:
        logging.debug("skipping extended arg")
    if library_keep is not None:
        adata = adata[:, list(map(lambda x: x == library_keep, adata.var["feature_types"]))].copy()
        logging.debug("filtered anndata size %s" % adata)
    
    return adata



def scp_read_10x_mtx(filename: PathLike, library_keep=None, *args, **kwargs) -> AnnData:
    """
    expanded sc.read_10x_mtx to filter for the library 
    adapted from https://github.com/scverse/muon/blob/master/muon/_prot/io.py
    """
    adata = read_10x_mtx(filename, gex_only=False, *args, **kwargs) #need to leave gex as this is scanpy's code
    logging.debug("filtering cellranger outputs to %s" % library_keep)
    if library_keep is not None:
        adata = adata[
            :, list(map(lambda x: x == library_keep, adata.var["feature_types"]))
        ].copy()
    return adata

#----------------
# temp loading functions for spatial data
# load_Spatial_in
# load_spatial_from_multiple_files
# ->>>how to write?
#----------------





def load_adata_in(path, filetype, gex_only=True, var_names="gene_symbols", library=None, extended=False):
    """
    load in any format supported by the load_functions_dict
    expected to be called by load_mdata_from_multiple_files
    """
    # all the supported loading functions here
    load_functions_dict = {
    "cellranger": lambda x: scp_read_10x_mtx(x, 
                                           library_keep=library,
                                           var_names=var_names,
                                           cache=False),
     "h5ad":read_h5ad,
     "csv_matrix":read_csv,
     "txt_matrix": read_text,
     "10X_h5": lambda x: scp_read_10x_h5(x, library_keep=library, extended=extended),
     "hd5": read_hdf,
      # vdj functions
     "cellranger_vdj": lambda x: read_scirpy(x, filetype="cellranger_vdj"),
     "tracer":lambda x: read_scirpy(x, filetype="tracer"),
     "bracer":lambda x: read_scirpy(x, filetype="bracer"),
     "airr":lambda x: read_scirpy(x, filetype="airr"),             
     }
    

    # the var_names argument decides which column of the cellranger features files ends up as the index.
    if var_names=="gene_symbols":
        col_fill="gene_ids"
    elif var_names=="gene_ids":
        col_fill="gene_symbols"
    else:
        col_fill="index"

    # try:
    logging.debug("loading %s " % filetype)
    adata = load_functions_dict[filetype](path)
    logging.debug("this is the anndata now: %s" % adata)
    # in some cases you need to update the var index col.
    if var_names in adata.var.columns:
        logging.info("resetting var.index")
        # we need to reset the index to be the expected one.
        adata.var = adata.var.reset_index().rename(columns={"index":col_fill}).set_index(var_names)
    return adata
    # except:
    #     raise ValueError("unknown filetype %s, please specify one of: %s" % filetype, ", ".join(list(load_functions_dict.keys()))) 
    


def update_intersecting_feature_names(adata1, adata2, prefix):
    # check for intersecting names
    if (len(adata1.var_names) > 0 and len(adata2.var_names) > 0 ):
        intersect_features = intersection(adata1.var_names.tolist(), adata2.var_names.tolist())
        if len(intersect_features) > 0:
            # this will be a problem downstream, so we prefix the 
            warnings.warn("intersecting gene ids and other modalities are problematic downstream, \
                prefixing  with %s" % prefix)
            adata2.var[ prefix + 'id'] = adata2.var_names
            adata2.var_names = [prefix + x for x in adata2.var_names.tolist()]
    return adata2


def merge_tcr_bcr_into_one_anndata(tcr, bcr):
    logging.info("merging tcr and bcr into one rep modality")
    intersect_obs = list(set(bcr.obs_names).union(set(tcr.obs_names)))
    adata = AnnData(X=np.empty(shape=(len(intersect_obs), 0)),
                       obs=pd.DataFrame(index=intersect_obs))
    # merge in the tcr and bcr with the rna 
    ir.pp.merge_airr_chains(adata , tcr)
    ir.pp.merge_airr_chains(adata , bcr)
    ir.tl.chain_qc(adata)
    return adata
    


def _make_one_rep_modality(data_dict):
    """
    takes the data dict from load_mdata_from_multiple_files, and collapses down the bcr and/or tcr into one modality called rep
    """
    if ("tcr" in data_dict.keys() and "bcr" in data_dict.keys()):
        # get the intersecting obs
        data_dict['rep'] = merge_tcr_bcr_into_one_anndata(data_dict['tcr'], data_dict['bcr'])
        # remove separate assays
        data_dict.pop('tcr', None)
        data_dict.pop('bcr', None)
    elif("tcr" in data_dict.keys()):
        logging.info("storing tcr in rep modality")
        data_dict['rep'] = data_dict['tcr']
        data_dict.pop('tcr', None)
    elif("bcr" in data_dict.keys()):
        logging.info("storing bcr in rep modality")
        data_dict['rep'] = data_dict['bcr']
        data_dict.pop('bcr', None)
    


def load_mdata_from_multiple_files(all_files_dict):
    """
    create a mudata object from multiples files
    Parameters
    ----------
    all_files_dict: dict
        dictionary containing one key per assay out of [RNA, PROT, TCR, BCR, ATAC]
        and the values for each key is a list of file path and fie type
        e.g.  {"RNA": [filepath, "filetype"],
               "PROT: [file path, "filetype"]} 
        Filetypes supported for RNA/prot: ["cellranger", "h5ad", "csv_matrix", "txt_matrix", "10X_h5"],
        Filetypes supported for atac (multiome preferred is 10X_h5) ["10X_h5","cellranger","h5ad"]
        Filetypes supported for rep: ["cellranger_vdj", "airr", "tracer", "bracer"  ] 
        See scirpy documentation for more information of repertoire input formats 
        https://scverse.org/scirpy/latest/api.html#module-scirpy.io
    """
    # convert names to match mudata conventions
    # mudata_conventional_names
    # {"GEX":"rna", "ADT":"prot", "TCR":"tcr", "BCR":"bcr", "ATAC": "atac", "spatial":"spatial"}
    # all_files_dict = {mudata_conventional_names[nm]: x  for (nm, x) in all_files_dict.items()}
    # note: scanpy's default function use gex_only as param so we need to leave that in
    logging.debug(all_files_dict.keys())
    # load in separate anndata for each expected modality
    data_dict = {}
    for nm, x in all_files_dict.items():
        logging.info("loading %s" % nm)
        extra_args={}
        if nm == "rna":
            extra_args['library'] = "Gene Expression"
            if 'atac' in all_files_dict.keys():
                extra_args["extended"]= True
            else:
                extra_args["extended"]= False
        if nm == "prot":
            extra_args["gex_only"] = False
            extra_args["var_names"] = "gene_ids"
            extra_args['library'] = "Antibody Capture"
            extra_args["extended"]= False
        if nm == "atac":
            extra_args["gex_only"] = False
            extra_args['library'] = "Peaks"
            extra_args['extended'] = False
        if nm == "spatial":
            extra_args["gex_only"] = True # check this for techs other than merfish and visium H&E
            #extra_args["counts_file"] =
            extra_args['extended'] = False

        logging.debug("extra args")
        logging.debug(extra_args)
        data_dict[nm] = load_adata_in(x[0], x[1], **extra_args) #** 
        #x[0] is the path, x[1] is the filetype
    logging.debug(data_dict["rna"])
    logging.debug(data_dict.keys())
    # we want unique var names for each assay
    if "rna" in data_dict.keys():
        logging.debug("rna adata %s " % data_dict["rna"])
        data_dict['rna'].var_names_make_unique()
        for mod in data_dict.keys():
            if mod =="rna":
                # skip since we want to update all the ones that are not rna
                pass
            else:
                logging.debug("second adata %s " % data_dict[mod])
                logging.debug("updating intersection for %s" % mod)
                data_dict[mod] = update_intersecting_feature_names(adata1=data_dict['rna'], adata2=data_dict[mod], 
                prefix=mod + "_")
    #make repertoire data into one assay
    _make_one_rep_modality(data_dict)
    logging.debug(data_dict)

    # finally make the mudata object
    mdata_ = mu.MuData(data_dict)
    mdata_.update()
    # remove empty_rna
    return mdata_


from scipy.io import mmwrite
import gzip
def write_10x_counts(adata, path, layer=None):
    if layer is None:
        arr = adata.X
    else:
        arr = adata.layers[layer]
    if os.path.exists(path) is False:
        os.makedirs(path)
    # convert to integer if appropriate
    if arr.sum().is_integer():
        arr = arr.astype(int)
    # write matrix.mtx
    mmwrite(os.path.join(path, "matrix.mtx"), arr.T)
    with open(os.path.join(path, "matrix.mtx"), 'rb') as f_in, gzip.open(os.path.join(path, "matrix.mtx.gz"), 'wb') as f_out:
        f_out.writelines(f_in)
    os.remove(os.path.join(path, "matrix.mtx"))
    # write barcodes file
    barcodes = adata.obs_names.to_frame()
    barcodes.to_csv(os.path.join(path, "barcodes.tsv.gz"), sep='\t', index=None, header=None)
    # write features file (trying to make sure the correct columns are stored)
    if "gene_ids" in adata.var.columns:
        # we assume the index is gene_ids
        features = adata.var.reset_index().rename(columns={"index": "gene_symbols"})
    elif "gene_symbols" in adata.var.columns:
        features = adata.var.reset_index().rename(columns={"index": "gene_ids"})
        if "prot_id" in features.columns:
            features = features.drop(columns=['gene_symbols']).rename(columns={"prot_id" : "gene_symbols"})
    features = features[['gene_ids','gene_symbols', 'feature_types']] 
    features.to_csv(os.path.join(path, "features.tsv.gz"), sep='\t', index=None, header=None)

# https://stackoverflow.com/questions/33529312/remove-empty-dicts-in-nested-dictionary-with-recursive-function
def dictionary_stripper(data):
    new_data = {}
    # Only iterate if the given dict is not None
    if data:
        for k, v in data.items():
            if isinstance(v, dict):
                v = dictionary_stripper(v)
            # ideally it should be not in, second you can also add a empty list if required
            if v not in ("", None, {}, []):
                new_data[k] = v
        # Only if you want the root dict to be None if empty
        if new_data == {}:
            return None
        return new_data
    return None


def replace_string_nones(d):
    for k, v in d.copy().items():
        if isinstance(v, dict):
            replace_string_nones(v)
        elif v == "None":
            d[k] = None

import yaml
def read_yaml(fname):
    if type(fname) is dict:
        params=fname
    elif os.path.exists(fname):
        with open(fname, "r") as stream:
            try:
                params = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
    else:
        # if the yaml is being passed as a string
        try:
            params = yaml.safe_load(fname)
            
        except yaml.YAMLError as exc:
            print(exc)
    replace_string_nones(params)
    return params


# def model_load_choice( ):
#     import scvi
#     functions_dict={
#     "totalvi": lambda x: scvi.model.TOTALVI.load_query_data,
#     "scvi": lambda x: scvi.model.SCVI.load_query_data,
#     "scanvi": lambda x: scvi.model.SCANVI.load_query_data,

#     }
    



'''
scanpy QC script ATAC
'''
from calendar import c
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import matplotlib.pyplot as plt
plt.ioff()
import numpy as np
import pandas as pd
import scanpy as sc
import anndata as ad
import os
import argparse
import seaborn as sns
from muon import atac as ac
import muon as mu

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("testing logger works")

from panpipes.funcs.io import  write_obs
from panpipes.funcs.processing import check_for_bool


sc.settings.verbosity = 3

parser = argparse.ArgumentParser()

parser.add_argument("--sampleprefix",
                    default="",
                    help="prefix to prepend when saving the metadata file")
parser.add_argument("--input_anndata",
                    default="adata_unfilt.h5ad",
                    help="")
parser.add_argument("--outfile",
                    default="adata_unfilt.h5ad",
                    help="")
parser.add_argument("--figdir",
                    default="./figures/",
                    help="path to save the figures to")
parser.add_argument("--figure_suffix",
                    default="_qc-plot.png",
                    help="figures filename suffix to be appended to figures/umap")
parser.add_argument('--is_paired',
                    default=True,
                    help='If ATAC data comes from a paired multiome experiment or is a standalone ATAC.')
parser.add_argument('--paired_rna',
                    default=None,
                    help='If ATAC data comes from standalone ATAC, do you have an rna anndata with obs.interval that you can use for Tss')
parser.add_argument('--tss_coordinates',
                    default=None,
                    help='If ATAC data comes from standalone ATAC, do you have a TSS annotation file? mm10 or hg19 supported')
parser.add_argument("--atac_qc_metrics",
                    default="",
                    help="comma sep list of params to quantify and plot")
parser.add_argument("--use_muon",
                    default="",
                    help="")
                



args, opt = parser.parse_known_args()

L.info("running with args:")
L.debug(args)

figdir = args.figdir

if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))
args.is_paired= check_for_bool(args.is_paired)

if args.is_paired:
    args.use_muon= True
    L.info("I'm working on a multiome experiment")
else:
    L.info("qc'ing an atac standalone assay")

mdata = mu.read(args.input_anndata)
atac = mdata.mod['atac']


# CALCULATE SOME GENERAL QC metrics
sc.pp.calculate_qc_metrics(atac, percent_top=None, log1p=False, inplace=True)
checkfiles = 'files' in list(atac.uns.keys()) 
if checkfiles:
    if atac.uns["files"]["fragments"] is not None:
        ac.tl.nucleosome_signal(atac, n=1e6)
        atac.obs['NS']=np.where(atac.obs['nucleosome_signal'] >3, 'trinucleosome', 'mono_trinucleosome')
        L.debug("added nucleosome signal to atac")
        mu.pl.histogram(atac, "nucleosome_signal", groupby='NS')
        plt.savefig(os.path.join(figdir, "nucleosome.png"))


if args.is_paired & checkfiles :
    if atac.uns["files"]["fragments"] is not None:
        #ac.tl.get_gene_annotation_from_rna(mdata['rna'])  # accepts MuData with 'rna' modality or mdata['rna'] AnnData directly if
        tss = ac.tl.tss_enrichment(mdata, n_tss=1000) # by default, features=ac.tl.get_gene_annotation_from_rna(mdata), but it could work with a custom annotation 
        tss.obs['tss_class'] = np.where(tss.obs["tss_score"]>2 , "High","Low") 
        ac.pl.tss_enrichment(tss, color="tss_class")
        plt.savefig(os.path.join(figdir, "tss_enrichment.png"))
        mdata.mod["tss"] = tss # add the tss anndata to the mudata, no point in saving it outside
        
else:
    #since we're here, change some names
    for col in atac.obs.columns:
        if "peak_region_fragments" in col:
            atac.obs["atac_peak_region_fragments"] = atac.obs[col]
        if col.endswith("mitochondrial"):
            if atac.obs["atac_mitochondrial_reads"] is not None:
                atac.obs["atac_mitochondrial_reads"] = atac.obs[col]
        if col.endswith("passed_filters"):  
            atac.obs['atac_fragments']= atac.obs[col]
    
    if args.paired_rna is not None:
        rna = mu.read(args.paired_rna)
        tss_features = ac.tl.get_gene_annotation_from_rna(rna)
    elif args.tss_coordinates is not None:
        tss_features = pd.read_csv(args.tss_coordinates, index_col=0, sep="\t")
    
        tss_data = ac.tl.tss_enrichment(atac, features= tss_features, n_tss=1000)
        tss_data.obs['tss_class'] = np.where(tss_data.obs["tss_score"]>2 , "High","Low") 
        ac.pl.tss_enrichment(tss_data, color="tss_class")
        plt.savefig(os.path.join(figdir, "tss_enrichment.png"))
        mdata.mod["tss"] = tss_data
    
if args.paired_rna is None :
    if args.tss_coordinates is None:
        L.warning("""you didn't provide an rna anndata or the features coordinate files to
        run the tss enrichment analysis, this will be skipped
        """)

# calc pct fragments in peaks 
if "atac_fragments" in atac.obs.columns:
    L.debug("adding percent fragments in peaks")
    atac.obs['pct_fragments_in_peaks']=atac.obs['atac_peak_region_fragments']/atac.obs['atac_fragments'] *100          

#PLOTS

if args.atac_qc_metrics is not None:
    qc_vars=args.atac_qc_metrics.split(",")
    qc_vars= [a.strip() for a in qc_vars]
else:
    qc_vars = ["n_genes_by_counts","total_counts","pct_fragments_in_peaks"]

qc_vars_plot = [gg for gg in qc_vars if gg in atac.obs.columns]
sc.pl.violin(atac, qc_vars_plot, 
             jitter=0.4, multi_panel=True, save = "atac_metrics_violin.png")

mdata.update()

L.info("saving mudata and obs in a metadata tsv file")
write_obs(mdata, output_prefix=args.sampleprefix, 
        output_suffix="_cell_metadata.tsv")

mdata.write(args.outfile)

L.info("Done")




import multiprocessing 

import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import argparse
import os
import muon as mu
from cgatcore import pipeline as P


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--input_mudata',
                    default='mudata_scaled.h5ad',
                    help='')
parser.add_argument('--use_muon',
                    default=False,
                    help='')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_totalvi.csv',
                    help='')
parser.add_argument('--integration_col', default='batch',
                    help='')
parser.add_argument('--figdir', default='./figures',
                    help='')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs", default=50)
parser.add_argument('--neighbors_k', default=30,
                    help="neighbors k")
parser.add_argument('--neighbors_metric',default="euclidean",
                    help="neighbor metric, e.g. euclidean or cosine")

args, opt = parser.parse_known_args()

# scanpy settings
sc.set_figure_params(facecolor="white")
sc.settings.autoshow = False
sc.settings.figdir = args.figdir


# load parameters
threads_available = multiprocessing.cpu_count()
params = P.get_parameters(
    ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
     "../pipeline.yml",
     "pipeline.yml"])


mdata = mu.read(args.input_mudata)
# we assume here there is a mudata object that already has a neighbours for each modality


mu.pp.neighbors(mdata)
mu.tl.umap(mdata)


L.info("Done")




#! /usr/bin/env python

from ruffus import *
import sys
import os
import re
from cgatcore import pipeline as P
import pandas as pd
import cgatcore.iotools as IOTools
from panpipes.funcs.io import check_submission_file, gen_load_anndata_jobs
# from scpipelines.funcs.processing import intersection
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning) 

import warnings

# warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
# warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)

import yaml


PARAMS = P.get_parameters(
    ["%s/pipeline.yml" % os.path.splitext(__file__)[0],
     "pipeline.yml"])

PARAMS['py_path'] =  os.path.join(os.path.dirname(os.path.dirname(__file__)), 'python_scripts')
PARAMS['r_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'R_scripts')
PARAMS['resources_path'] = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")
job_kwargs = {}

if PARAMS['condaenv'] is not None:
    job_kwargs["job_condaenv"] =PARAMS['condaenv']


#------------------------------------------------------------------------------------------------
## Create a dictionary of modalities
#------------------------------------------------------------------------------------------------
mode_dictionary = PARAMS["modalities"]
#{'rna': True, 'prot': False, 'Repertoire': False, 'atac': False}

# -----------------------------------------------------------------------------------------------
## 10X metrics plotting
# -----------------------------------------------------------------------------------------------

# TODO: update this to check for each modality csv file
# TODO: check this works with current submission file
@active_if(PARAMS['plot_10X_metrics'])
@follows(mkdir('figures'))
@follows(mkdir('figures/tenx_metrics'))
@follows(mkdir("logs"))
@originate("logs/tenx_metrics_multi_aggregate.log")
def aggregate_tenx_metrics_multi(outfile):
    """this is to aggregate all the cellranger multi metric_summary files
    it also does some plotting
    """    
    cmd = """
        python %(py_path)s/aggregate_cellranger_summary_metrics.py
            --pipe_df %(submission_file)s
            --figdir figures/tenx_metrics/
            --cellranger_column_conversion_df %(resources_path)s/metrics_summary_col_conversion.tsv
            --output_file 10x_metrics.csv > %(outfile)s
            """
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)


@follows(aggregate_tenx_metrics_multi)
def process_all_tenx_metrics():
    pass
# -----------------------------------------------------------------------------------------------
## Creating h5mu from filtered data files
# -----------------------------------------------------------------------------------------------

def unfilt_file():
    sprefix = PARAMS['sample_prefix']
    unfilt_file = sprefix + "_unfilt.h5mu"
    return unfilt_file



def gen_load_filtered_anndata_jobs():
    caf = pd.read_csv(PARAMS['submission_file'], sep='\t')
    return gen_load_anndata_jobs(caf, load_raw=False, mode_dictionary=PARAMS["modalities"], load_prot_from_raw=PARAMS['load_prot_from_raw'])

    

@follows(mkdir("logs"))
@follows(mkdir("tmp"))
@active_if(PARAMS["use_existing_h5mu"] is False)
@files(gen_load_filtered_anndata_jobs)
def load_mudatas(rna_path, outfile, 
              sample_id, 
              rna_filetype,  
              prot_path, prot_filetype, 
              tcr_path, tcr_filetype,  
              bcr_path, bcr_filetype, 
              atac_path, atac_filetype, 
              fragments_file, per_barcode_metrics_file, peak_annotation_file, 
              cell_mtd_path):
    
    path_dict = {'rna':rna_path,
                 'prot': prot_path,
                 'bcr': bcr_path,
                 'tcr' : tcr_path,
                 'atac':atac_path}
    print(path_dict)
    
    modality_dict = {k:True if path_dict[k] is not None else False for k,v in PARAMS['modalities'].items() }
    print(modality_dict)
    
    cmd = """
        python %(py_path)s/make_adata_from_csv.py 
        --mode_dictionary "%(modality_dict)s"
        --sample_id %(sample_id)s
        --output_file %(outfile)s 
    """
    
    cmd += " --use_muon True" #this is still here?
    if rna_path is not None and pd.notna(rna_path):
        cmd += " --rna_infile %(rna_path)s"
        cmd += " --rna_filetype %(rna_filetype)s"
    if prot_path is not None and pd.notna(prot_path):
        cmd += " --prot_infile %(prot_path)s"
        cmd += " --prot_filetype %(prot_filetype)s"
        cmd += " --subset_prot_barcodes_to_rna %(subset_prot_barcodes_to_rna)s"
    if atac_path is not None and pd.notna(atac_path):
        cmd += " --atac_infile %(atac_path)s"
        cmd += " --atac_filetype %(atac_filetype)s"
    if per_barcode_metrics_file is not None and pd.notna(per_barcode_metrics_file):
        cmd += " --per_barcode_metrics_file %(per_barcode_metrics_file)s"
    if fragments_file is not None and pd.notna(fragments_file):
        cmd += " --fragments_file %(fragments_file)s"
    if peak_annotation_file is not None and pd.notna(peak_annotation_file):
        cmd += " --peak_annotation_file %(peak_annotation_file)s"
      # ~ means this tests "is not nan"
    if tcr_path is not None and pd.notna(tcr_path):
        cmd += " --tcr_filtered_contigs %(tcr_path)s"
        cmd += " --tcr_filetype %(tcr_filetype)s"
    if bcr_path is not None and pd.notna(bcr_path):
        cmd += " --bcr_filtered_contigs %(bcr_path)s"
        cmd += " --bcr_filetype %(bcr_filetype)s"
    cmd += " > logs/load_mudatas_%(sample_id)s.log"
    # print(cmd)
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)



@active_if(PARAMS["use_existing_h5mu"] is False)
@collate(load_mudatas,
         formatter(""),
         unfilt_file())
def concat_filtered_mudatas(infiles, outfile):
    # print(infiles)
    # print(outfile)
    infiles_str = ','.join(infiles)
    cmd = """
    python %(py_path)s/concat_adata.py 
        --input_files_str %(infiles_str)s 
        --output_file %(outfile)s 
        --submissionfile %(submission_file)s
        --sampleprefix %(sample_prefix)s
        --join_type %(concat_join_type)s
        """
    if PARAMS['metadatacols'] is not None and PARAMS['metadatacols'] != "":
        cmd += " --metadatacols  %(metadatacols)s"
    if PARAMS["barcode_mtd_include"] is True:
        cmd += " --barcode_mtd_df %(barcode_mtd_path)s"
        cmd += " --barcode_mtd_metadatacols %(barcode_mtd_metadatacols)s"
    if PARAMS['protein_metadata_table'] is not None:
        cmd += " --protein_var_table %(protein_metadata_table)s"
    if PARAMS['index_col_choice'] is not None:
        cmd += " --protein_new_index_col %(index_col_choice)s"
    cmd += " > logs/concat_filtered_mudatas.log"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)
    # P.run("rm tmp/*", job_threads=PARAMS['resources_threads_low'])


# -----------------------------------------------------------------------------------------------
## Creating h5mu from bg data files
# -----------------------------------------------------------------------------------------------

if PARAMS['assess_background'] or (PARAMS["modalities"]["prot"] and "dsb" in PARAMS['normalisation_methods']):
    PARAMS['bg_required'] = True
else:
    PARAMS['bg_required'] = False
    

def bg_file():
    sprefix = PARAMS['sample_prefix']
    use_muon = True
    if use_muon:
        bg_file = sprefix + "_bg.h5mu"
    else:
        bg_file = sprefix + "_bg.h5ad"
    return bg_file


def gen_load_bg_anndata_jobs():
    caf = pd.read_csv(PARAMS['submission_file'], sep='\t')
    return gen_load_anndata_jobs(caf, load_raw=True, mode_dictionary=PARAMS["modalities"], load_prot_from_raw=True)
    

@active_if(PARAMS["bg_required"])
@follows(mkdir("logs"))
@follows(mkdir("tmp"))
@active_if(PARAMS["use_existing_h5mu"] is False)
@files(gen_load_bg_anndata_jobs)
def load_bg_mudatas(rna_path, outfile, 
              sample_id, 
              rna_filetype,  
              prot_path, prot_filetype, 
              tcr_path, tcr_filetype,  
              bcr_path, bcr_filetype, 
              atac_path, atac_filetype, 
              fragments_file, per_barcode_metrics_file, peak_annotation_file, 
              cell_mtd_path):
    path_dict = {'rna':rna_path,
                 'prot': prot_path,
                 'bcr': bcr_path,
                 'tcr' : tcr_path,
                 'atac':atac_path}
    print(path_dict)
    
    # need to remove unnessacry modality
    modality_dict = {k:True if path_dict[k] is not None else False for k,v in PARAMS['modalities'].items() }

    modality_dict['tcr'] = False
    modality_dict['bcr'] = False
    cmd = """
    python %(py_path)s/make_adata_from_csv.py 
    --sample_id %(sample_id)s
    --mode_dictionary "%(modality_dict)s"
    --rna_infile %(rna_path)s
    --rna_filetype %(rna_filetype)s
    --output_file %(outfile)s 
    """
    if prot_path is not None and pd.notna(prot_path):
        cmd += " --prot_infile %(prot_path)s"
        cmd += " --prot_filetype %(prot_filetype)s"
    if atac_path is not None and pd.notna(atac_path):
        cmd += " --atac_infile %(atac_path)s"
        cmd += " --atac_filetype %(atac_filetype)s"
    if per_barcode_metrics_file is not None and pd.notna(per_barcode_metrics_file):
        cmd += " --per_barcode_metrics_file %(per_barcode_metrics_file)s"
    if fragments_file is not None and pd.notna(fragments_file):
        cmd += " --fragments_file %(fragments_file)s"
    if peak_annotation_file is not None and pd.notna(peak_annotation_file):
        cmd += " --peak_annotation_file %(peak_annotation_file)s"
    if PARAMS['protein_metadata_table'] is not None:
        cmd += " --protein_var_table %(protein_metadata_table)s"  #check which of these 2 needs to stay!!!
    if PARAMS['index_col_choice'] is not None:
        cmd += " --protein_new_index_col %(index_col_choice)s"
    cmd += " > logs/load_bg_mudatas_%(sample_id)s.log"
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)


@active_if(PARAMS["bg_required"])
@active_if(PARAMS['downsample_background'])
@follows(load_bg_mudatas)
@transform(load_bg_mudatas, 
           regex("./tmp/(.*)_raw.h5(.*)"), 
           r"./logs/\1_bg_downsampled.log")
def downsample_bg_mudatas(infile, outfile):
    cmd = """
    python %(py_path)s/downsample.py 
    --input_mudata %(infile)s
    --output_mudata "%(infile)s" 
    --downsample_value 20000  > %(outfile)s
    """
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd, **job_kwargs)


@active_if(PARAMS["bg_required"])
@active_if(PARAMS["assess_background"])
@active_if(PARAMS["use_existing_h5mu"] is False)
@follows(load_bg_mudatas)
@follows(downsample_bg_mudatas)
@collate(load_bg_mudatas,
         formatter(""),
         bg_file())
def concat_bg_mudatas(infiles, outfile):
    # print(infiles)
    # print(outfile)
    infiles_str = ','.join(infiles)
    cmd = """
    python %(py_path)s/concat_adata.py 
        --input_files_str %(infiles_str)s 
        --output_file %(outfile)s 
        --submissionfile %(submission_file)s
        --sampleprefix %(sample_prefix)s
        --join_type %(concat_join_type)s
        """
    if PARAMS['metadatacols'] is not None and PARAMS['metadatacols'] != "":
        cmd += " --metadatacols  %(metadatacols)s"
  #  if PARAMS["barcode_mtd_include"] is True:
   #     cmd += " --barcode_mtd_df %(barcode_mtd_path)s"
    #    cmd += " --barcode_mtd_metadatacols %(barcode_mtd_metadatacols)s"
    cmd += " > logs/concat_bg_mudatas.log"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)
    # P.run("rm tmp/*", job_threads=PARAMS['resources_threads_low'])

# -----------------------------------------------------------------------------------------------
## rna QC 
# -----------------------------------------------------------------------------------------------


def orfile():
    return PARAMS['sample_prefix'] + "_cell_metadata.tsv"
@active_if(PARAMS["scr_run"])
@active_if(PARAMS['modalities_rna'])
@active_if(PARAMS["use_existing_h5mu"] is False)
@follows(mkdir("scrublet"))
@transform(load_mudatas, regex(r"./tmp/(.*).h5(.*)"),
           r"scrublet/\1_scrublet_scores.txt", r"\1")
def run_scrublet(infile, outfile, sample_id):
    outdir = "./scrublet"
    cmd = """python %(py_path)s/run_scrublet_scores.py
        --sample_id %(sample_id)s
        --inputpath %(infile)s
        --outdir %(outdir)s
        """
    if PARAMS['scr_expected_doublet_rate'] is not None:
        cmd += " --expected_doublet_rate %(scr_expected_doublet_rate)s"
    if PARAMS['scr_sim_doublet_ratio'] is not None:
        cmd += " --sim_doublet_ratio %(scr_sim_doublet_ratio)s"
    if PARAMS['scr_n_neighbours'] is not None:
        cmd += " --n_neighbors %(scr_n_neighbours)s"
    if PARAMS['scr_min_counts'] is not None:
        cmd += " --min_counts %(scr_min_counts)s"
    if PARAMS['scr_min_cells'] is not None:
        cmd += " --min_cells %(scr_min_cells)s"
    if PARAMS['scr_min_gene_variability_pctl'] is not None:
        cmd += " --min_gene_variability_pctl %(scr_min_gene_variability_pctl)s"
    if PARAMS['scr_n_prin_comps'] is not None:
        cmd += " --n_prin_comps %(scr_n_prin_comps)s"
    if PARAMS['scr_use_thr'] is not None:
        cmd += " --use_thr %(scr_use_thr)s"
    if PARAMS['scr_call_doublets_thr'] is not None:
        cmd += " --call_doublets_thr %(scr_call_doublets_thr)s"
    cmd += " > logs/run_scrublet_" + sample_id + ".log"
    job_kwargs["job_threads"] = PARAMS['resources_threads_medium']
    P.run(cmd,**job_kwargs)
    IOTools.touch_file(outfile)


@active_if(PARAMS['modalities_rna'])
@follows(mkdir("figures"))
@follows(mkdir("figures/rna"))
@follows(concat_filtered_mudatas, run_scrublet)
@originate("logs/run_scanpy_qc_rna.log", orfile(), unfilt_file())
def run_rna_qc(log_file, outfile, unfilt_file):
    # infile = submission file
    resources_path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources")
    customgenesfile = PARAMS["custom_genes_file"]
    calcproportions= PARAMS["calc_proportions"]
    cmd = """
        python %(py_path)s/run_scanpyQC_rna.py
          --sampleprefix %(sample_prefix)s
          --input_anndata %(unfilt_file)s
          --outfile %(unfilt_file)s
          --customgenesfile %(customgenesfile)s
          --calc_proportions %(calcproportions)s
          --figdir figures
          """
    if len(os.listdir('scrublet')) != 0:
        cmd += " --scrubletdir scrublet"
    # if params is specified otherwise default values will be used.
    # if params is specified otherwise default values will be used.
    if PARAMS['score_genes'] is not None:
        scoregenes = PARAMS['score_genes']
        cmd += " --score_genes %(scoregenes)s"
    if PARAMS['ccgenes'] is not None:
        if PARAMS['ccgenes'] == "default":
            ccgenesfile = resources_path + "/cell_cycle_genes.tsv"
        else:
            ccgenesfile = PARAMS['ccgenes']
        cmd += " --ccgenes %(ccgenesfile)s"
    cmd += " --channel_col %(channel_col)s" # is this needed?

    # if PARAMS['isotype_upper_quantile'] is not None:
    #     cmd += " --isotype_upper_quantile %(isotype_upper_quantile)s"
    # if PARAMS['isotype_n_pass'] is not None:
    #     cmd += " --isotype_n_pass %(isotype_n_pass)s"
    # add log file
    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)
    if os.path.exists("cache"):
        P.run("rm -r cache")
    #IOTools.touch_file(metadata)

#TO DO ADD the other modalities in this order:

# prot run_scanpy_qc_prot.log


# -----------------------------------------------------------------------------------------------
## prot QC 
# -----------------------------------------------------------------------------------------------
    
# this is because we don't want to try and load the file during the `config` stage

# def prot_cell_mtd():
#     return PARAMS['sample_prefix'] + "_prot_cell_metadata.tsv"
@active_if(PARAMS['modalities_prot'])
@follows(mkdir('figures/prot'))
@follows(concat_filtered_mudatas, run_rna_qc)
@originate("logs/run_scanpy_qc_prot.log", orfile(), unfilt_file())
def run_scanpy_prot_qc(log_file, outfile, unfilt_file):
    # infile = submission file
    cmd = """
        python %(py_path)s/run_scanpyQC_prot.py
          --sampleprefix %(sample_prefix)s
          --input_anndata %(unfilt_file)s
          --outfile %(unfilt_file)s
          --figdir figures/prot/
          """
    if PARAMS['channel_col'] is not None:
        cmd += " --channel_col %(channel_col)s"
    else:
        cmd += " --channel_col sample_id"
    if PARAMS['plotqc_prot_metrics']:
        cmd += " --per_cell_metrics %(plotqc_prot_metrics)s"
    if PARAMS['plot_metrics_per_prot']:
        cmd += " --per_prot_metrics %(plot_metrics_per_prot)s"
    if PARAMS['identify_isotype_outliers']:
        cmd += " --identify_isotype_outliers %(identify_isotype_outliers)s"
        cmd += " --isotype_upper_quantile %(isotype_upper_quantile)s"
        cmd += " --isotype_n_pass %(isotype_n_pass)s"
    # add log file
    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)
    pass

@active_if(PARAMS['modalities_prot'])
@follows(run_scanpy_prot_qc, concat_filtered_mudatas, concat_bg_mudatas)
@originate("logs/run_dsb_clr.log", unfilt_file(), bg_file())
def run_dsb_clr(outfile, unfilt_file, bg_file):
    print(unfilt_file)
    # infile = mdata_unfilt
    # outfile sampleprefix + "prot_qc_per_cell.tsv"
    cmd = """
        python %(py_path)s/run_preprocess_prot.py
        --filtered_mudata %(unfilt_file)s
        --figpath ./figures/prot
        """
    if PARAMS['channel_col'] is not None:
        cmd += " --channel_col %(channel_col)s"
    if PARAMS['normalisation_methods'] is not None:
        cmd += " --normalisation_methods %(normalisation_methods)s"
    if PARAMS['quantile_clipping'] is not None:
        cmd += " --quantile_clipping %(quantile_clipping)s"
    if PARAMS['clr_margin'] is not None:
        cmd += " --clr_margin %(clr_margin)s"
    if PARAMS['save_norm_prot_mtx'] is True:
        cmd += " --save_mtx True"
    # find the bg data if available
    if os.path.exists(bg_file):
        cmd += " --bg_mudata %(bg_file)s"
    cmd += " > %(outfile)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)

@follows(run_scanpy_prot_qc, run_dsb_clr)
def run_prot_qc():
    pass


# -----------------------------------------------------------------------------------------------
## Repertoire QC  # Repertoire run_scanpy_qc_rep.log
# -----------------------------------------------------------------------------------------------
# toggle_rep_qc = ()
@active_if(PARAMS['modalities_bcr'] or PARAMS['modalities_tcr']  )
@follows(mkdir('figures/rep'))
@follows(run_rna_qc, run_prot_qc)
@originate("logs/run_scanpy_qc_rep.log", unfilt_file())
def run_repertoire_qc(logfile, unfilt_file):
    cmd = """python %(py_path)s/run_scanpyQC_rep.py
          --sampleprefix %(sample_prefix)s
          --input_mudata %(unfilt_file)s
          --output_mudata %(unfilt_file)s
          --figdir figures/rep/
          --distance_metrics "%(ir_dist)s"
          --clonotype_metrics "%(clonotype_definition)s"
          """
    cmd += " > %(logfile)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)

# -----------------------------------------------------------------------------------------------
## atac QC # run_scanpy_qc_atac.log
# -----------------------------------------------------------------------------------------------
@active_if(PARAMS['modalities_atac'])
@mkdir('figures/atac')
@follows(run_rna_qc, run_prot_qc, run_repertoire_qc)
@originate("logs/run_scanpy_qc_atac.log", orfile(), unfilt_file())
def run_atac_qc(log_file, outfile, unfilt_file):
    # if this is a multiple samples project
    # they should be run together upfront 
    # for independent atac run cellranger-atac count, then cellranger-atac aggr
    # https://support.10xgenomics.com/single-cell-atac/software/pipelines/latest/using/aggr
    # for multiome run cellranger-arc count then cellranger-arc aggr
    # https://support.10xgenomics.com/single-cell-multiome-atac-gex/software/pipelines/latest/using/aggr
    # 
    

    plot_qc_atac_metrics = PARAMS["plotqc_atac_metrics"]
    cmd = """
        python %(py_path)s/run_scanpyQC_atac.py
          --sampleprefix %(sample_prefix)s
          --input_anndata %(unfilt_file)s
          --outfile %(unfilt_file)s
          --figdir figures/atac/  
          --atac_qc_metrics %(plot_qc_atac_metrics)s
          """
    if PARAMS["is_paired"]:
        cmd += " --is_paired True"
    else:
        cmd += " --is_paired False"
    if PARAMS["features_tss"] is not None:
        cmd += " --is_paired False"
        feat_file=PARAMS["features_tss"]
        cmd += " --tss_coordinates %(feat_file)s"
    if PARAMS["partner_rna"] is not None:
        cmd += " --is_paired False"
        prna_file = PARAMS["partner_rna"]
        cmd += " --paired_rna %(prna_file)s"
    
    cmd += " --use_muon True"

    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)

@follows(run_rna_qc, run_prot_qc, run_repertoire_qc, run_atac_qc)
def run_qc():
    pass


# -----------------------------------------------------------------------------------------------
## QC plotting
# -----------------------------------------------------------------------------------------------

@follows(run_qc)
@originate("logs/plot_qc.log", orfile())
def plot_qc(log_file, cell_file):
    qcmetrics = PARAMS['plotqc_rna_metrics']
    cmd = """
    Rscript %(r_path)s/plotQC.R 
    --prefilter TRUE
    --cell_metadata %(cell_file)s 
    --sampleprefix %(sample_prefix)s
    --groupingvar %(plotqc_grouping_var)s
    """
    cmd += " --scanpy_or_muon muon"
    if PARAMS["modalities"]['rna'] and PARAMS['plotqc_rna_metrics'] is not None:
        cmd += " --rna_qc_metrics %(plotqc_rna_metrics)s"
    if PARAMS["modalities"]['prot'] and PARAMS['plotqc_prot_metrics'] is not None:
        cmd += " --prot_qc_metrics %(plotqc_prot_metrics)s"
    if PARAMS["modalities"]['atac'] and PARAMS['plotqc_atac_metrics'] is not None:
        cmd += " --atac_qc_metrics %(plotqc_atac_metrics)s"
    if PARAMS['modalities_bcr'] or PARAMS['modalities_tcr']:
        if PARAMS['plotqc_rep_metrics'] is not None:
            pqrm = ','.join(['plotqc_rep_metrics'])
            cmd += " --rep_qc_metrics %(pqrm)s"
    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_low']
    P.run(cmd, **job_kwargs)


# ------
# assess background
# -------

@active_if(PARAMS['assess_background'])
@follows(mkdir("figures/background"))
@follows(run_qc)
@follows(concat_bg_mudatas)
@originate("logs/assess_background.log", unfilt_file(), bg_file())
def run_assess_background(log_file, unfilt_file, bg_file):
    cmd = """
    python %(py_path)s/assess_background.py
        --filtered_mudata %(unfilt_file)s
        --bg_mudata %(bg_file)s
        --channel_col %(channel_col)s
        --figpath "./figures/background/"
    """
    cmd += " > %(log_file)s"
    job_kwargs["job_threads"] = PARAMS['resources_threads_high']
    P.run(cmd, **job_kwargs)



# # ------------
# TO DO change the decorator to follow all qc plots?
@follows(plot_qc)
@follows(run_rna_qc, run_assess_background)
def all_rna_qc():
    pass

@follows(run_dsb_clr, plot_qc)
def all_prot_qc():
    pass

@follows(process_all_tenx_metrics)
@follows(all_prot_qc)
@follows(all_rna_qc)
#@originate("ruffus.check")
def full():
    """
    All cgat pipelines should end with a full() function which updates,
    if needed, all branches of the pipeline.
    The @follows statement should ensure that all functions are covered,
    either directly or as prerequisites.
    """
    #IOTools.touch_file(file)
    pass

@originate("cleanup_done.txt")
def cleanup(file):
    # remove any ctmp fails
    P.run("rm ctmp*", without_cluster=True)
    # remove tmp dir
    P.run("rm -r tmp", without_cluster=True)
    # delete empty dirs
    P.run("find ./ -empty -type d -delete", without_cluster=True)

def main(argv=None):
    if argv is None:
        argv = sys.argv
    P.main(argv)


if __name__ == "__main__":
    sys.exit(P.main(sys.argv))



import scanpy as sc
import argparse
import pandas as pd
import re
import numpy as np
from scipy.sparse import issparse
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata
from panpipes.funcs.scmethods import pseudo_seurat

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--infile",
                    default="rnapbmc3k_nn30_drharmony_metcosine_neighbors.h5ad",
                    help="file name, format: .h5ad")
parser.add_argument('--use_muon',
                    default=False,
                    help='')
parser.add_argument('--modality',
                    default="rna",
                    help='')
parser.add_argument("--outputfile",
                    default="rnacluster1_markers.txt",
                    help="file name, format: .txt")
parser.add_argument("--cluster_file",
                    default="rnapbmc3k_nn30_res0.2_algleiden_cluster.txt",
                    help="file name, format: .h5ad")
parser.add_argument("--cluster_value", type=str,
                    default='1', 
                    help="the ID of the cluster that we want to look for markers in.")
parser.add_argument("--testuse", 
                    default='wilcoxon', 
                    help="wilcoxon, t-test or logreg")
parser.add_argument("--pseudo_seurat",
                    default=False, type=check_for_bool,
                    help="apply seurat like filtering before marker test")
parser.add_argument("--minpct", 
                    default=0.1,
                    help="minimum fraction of cells expressing gene")
parser.add_argument("--mindiffpct", type=str, 
                    default='-inf',
                    help="minimum fraction difference between cluster and other cells. Setting not recommended.")
parser.add_argument("--threshuse",  
                    default=0.25,
                    help="testing limited to genes with this (log scale) difference in mean expression level.")
parser.add_argument("--mincells", 
                    default=3,
                    help="minimum number of cells required (applies to cluster and to the other cells)")


args, opt = parser.parse_known_args()
use_muon = args.use_muon


# script -------------------------------

# check for options

L.info("Running with options")
L.info(args)
L.info('\n')
# read data
adata = read_anndata(args.infile, use_muon=use_muon, modality=args.modality)

# adata = sc.read_h5ad("../../data/scanpy_pipe_test/pbmc3k_nn30_drharmony_metcosine_neighbors.h5ad")

# load clusters
df = pd.read_csv(args.cluster_file, sep="\t", index_col=0)
# df = pd.read_csv("../../data/scanpy_pipe_test/pbmc3k_res0.6_nn30_algleiden_cluster.txt", sep="\t", index_col=0 )
#
adata_shape = adata.shape


# add clusters to adata.obs
adata.obs = pd.merge(adata.obs, df, how="left", left_index=True, right_index=True )
# check check we have not lost cells in merge
if adata.shape != adata_shape:
    L.info("some cells lost in merge, not all cells have a cluster?")



# how many unique clusters?

# 1. set idents
adata.obs["idents"] = ["1" if cc==int(args.cluster_value) else "0" for cc in adata.obs["clusters"]]
adata.obs["idents"] = adata.obs["idents"].astype("category")

# check we have enough cells
if args.mincells == "None":
    min_cells = 3
else:
    min_cells = int(args.mincells)

n_cluster = sum([cell is '1' for cell in adata.obs['idents']])
n_other = sum([cell is '0' for cell in adata.obs['idents']])
if n_cluster <= int(args.mincells) | n_other <= int(args.mincells):
    sys.exit("not enough cells in cluster, min cells: %i" % int(args.mincells))

L.info("Cluster %s number of cells: %i" % (args.cluster_value, n_cluster))
L.info("Other number of cells: %i\n" % (n_other))

# there is some weird difference between dense and sparse array?
if issparse(adata.X):
    bool_list = (adata.X.sum(axis=0) > 0).tolist()[0]
else:
    bool_list = (adata.X.sum(axis=0) > 0).tolist()
adata = adata[:, bool_list]

if isinstance(args.pseudo_seurat, str):
    if args.pseudo_seurat == 'True':
        run_pseudo_seurat = True
    elif args.pseudo_seurat == 'False':
        run_pseudo_seurat = False
    else:
        L.info(args.pseudo_seurat)
        sys.exit("spelling_error: please specify --pseudoseurat as True or False")
elif type(args.pseudo_seurat) == 'bool':
    run_pseudo_seurat = args.pseudo_seurat
else:
    sys.exit("type error %s: please specify --pseudo_seurat as True or False" % type(args.pseudo_seurat))


if run_pseudo_seurat:
    L.info("running pseudo seurat with params: \nminpct: %s \nmindiffpct: %s \nlogFCthreshold: %s\n\n" % (args.minpct, args.mindiffpct, args.threshuse))

    filter_stats = pseudo_seurat(adata,
                                 arg_minpct=float(args.minpct),
                                 arg_mindiffpct=float(args.mindiffpct),
                                 arg_logfcdiff=float(args.threshuse))
    L.info("number of genes remaining after filtering:  %i\n" % filter_stats['background'].sum())
    # adata = adata.raw.to_adata().copy()
    adata = adata[:, filter_stats['background'].tolist()]
#     adata.raw = adata
    L.info(adata)
    L.info('\n')
else:
    # get backgrounds
    pass
    # write out background after rank_gene_groups


# now we can actually run find markers,
# run on a subset of genes
sc.tl.rank_genes_groups(adata, groups=["1"], groupby="idents", reference="0",
                        method="wilcoxon", n_genes=float("inf"), corr_method="bonferroni", layer="logged_counts")

markers = sc.get.rank_genes_groups_df(adata, group="1")

# merge markers with filter_stats_df, so that we can see all the scroes for gsea
#

# filter positive only and adjusted pval < 0.05
# markers = markers[(markers.logfoldchanges > 0) & (markers.pvals_adj < 0.05)]

if sc.__version__ < "1.7.0":
    markers.columns = ['scores', 'gene', 'avg_logFC', 'pvals', 'p.adj.bonferroni']
elif sc.__version__ == "1.7.0":
    markers.columns = ['gene', 'scores', 'avg_logFC', 'pvals', 'p.adj.bonferroni']

markers.head()
L.info("Number of Markers for cluster %s: %i" % (args.cluster_value, markers.shape[0]))
L.info("Rank gene groups complete")


L.info("saving markers to:" + args.outputfile)
markers.to_csv(args.outputfile, sep="\t", index=False)

if run_pseudo_seurat:
    fname = re.sub('_markers.txt', '_background.txt', args.outputfile)
    L.info("saving background to:" + fname)
    filter_stats.to_csv(fname, sep="\t", index=False)


# quick do some plotting!

L.info("Done")




import argparse
import pandas as pd
import os
import numpy as np
import scanpy as sc
import muon as mu
from matplotlib import use
use('Agg')
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use("Agg")
import seaborn as sns
import panpipes.funcs as pnp
from scipy.stats import rankdata

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)
L.debug("test logging message")
# parse arguments
parser = argparse.ArgumentParser()


parser.add_argument('--channel_col',
                    default="sample_id",
                    help='')
parser.add_argument('--filtered_mudata',
                    default=None,
                    help='')  
parser.add_argument('--bg_mudata',
                    default=None,
                    help='')    
parser.add_argument('--figpath',
                    default="./figures/background",
                    help='')  


args, opt = parser.parse_known_args()
L.info(args)

sc.settings.figdir = args.figpath
# load filtered data - if use_umon is True this will return a mudata object, else returns an mudata object

L.info("reading filtered mudata object")
try:
    mdata = mu.read(args.filtered_mudata)
    # need to delete rep else we lose too many cells in intersect_obs later
    if "rep" in mdata.mod.keys():
        del mdata.mod["rep"]
except FileNotFoundError:
    sys.exit("filtered mudata object not found")


# mu.pp.intersect_obs(mdata)
# load bg data

L.info("reading bg mudata object")
try:
    mdata_bg = mu.read(args.bg_mudata)
except FileNotFoundError:
    sys.exit("bg_mudata object not found")

# barcode rank plots
df = mdata_bg['rna'].obs.copy()
df['total_counts'] = np.ravel(mdata_bg['rna'].X.sum(axis=1))
df['log_total_counts'] = np.log(df['total_counts'])
df['ranks'] = df.groupby('sample_id', as_index=False)['total_counts'].transform(lambda x: rankdata(x*-1, method="ordinal"))
df['log_ranks'] = np.log(df['ranks'])
df['is_cell'] = df.index.isin(mdata['rna'].obs_names.tolist())
df['is_cell'] = df['is_cell'].astype('category')
df['is_cell'] = df['is_cell'].cat.set_categories([True, False])

fig, ax= plt.subplots(figsize=(6,6))
sns.scatterplot(data=df, x="log_ranks", y="log_total_counts", 
                hue=args.channel_col, size="is_cell",  sizes=(2, 20),
                ax=ax, linewidth=0)
plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
plt.savefig(os.path.join(args.figpath,'rna_barcode_ranks.png'), bbox_inches="tight")
plt.clf()

n_samples = len(df[args.channel_col].unique())
if n_samples > 1:
    g = sns.FacetGrid(df, col=args.channel_col, 
                    height=2.5,  
                    col_wrap=np.min([n_samples,8]))
    g.map(sns.scatterplot, "log_ranks", "log_total_counts", linewidth=0, s=1.5)
    g.set_titles(col_template="{col_name}")
    plt.savefig(os.path.join(args.figpath,'rna_facet_barcode_ranks.png'))
    plt.clf()



if 'rna' in mdata.mod.keys():
    # remove cells that passed cellranger filtering from bg
    mdata_bg['rna'].obs['is_empty'] = ~mdata_bg['rna'].obs_names.isin(mdata['rna'].obs_names)
    mu.pp.filter_obs(mdata_bg['rna'], 'is_empty')

# mdata_bg.update()
# mu.pp.intersect_obs(mdata_bg)

# calculate qc metrics
if 'rna' in mdata.mod.keys():
    # run stadard qc metrics
    sc.pp.calculate_qc_metrics(mdata['rna'], inplace=True, percent_top=None, log1p=True)
    sc.pp.calculate_qc_metrics(mdata_bg['rna'], inplace=True, percent_top=None, log1p=True)



if 'prot' in mdata.mod.keys():
    sc.pp.calculate_qc_metrics(mdata['prot'], inplace=True, percent_top=None, log1p=True)
    sc.pp.calculate_qc_metrics(mdata_bg['prot'], inplace=True, percent_top=None, log1p=True)


mdata.update()
mdata_bg.update()

if 'rna' in mdata.mod.keys():
    n_samples_rna=len(mdata['rna'].obs['sample_id'].unique())
if "prot" in mdata.mod.keys():
    n_samples_prot=len(mdata['prot'].obs['sample_id'].unique())

# quantifying the top background features
if os.path.exists(args.figpath) is False:
    os.mkdirs(args.figpath)

if 'rna' in mdata.mod.keys():
    sc.pl.highest_expr_genes(mdata_bg['rna'],n_top=30, save="_rna_background.png")
    top_genes = pnp.scmethods.get_top_expressed_features(mdata_bg['rna'], n_top=30, group_by=args.channel_col)
    bg_df = pnp.scmethods.get_mean_background_fraction(mdata_bg['rna'], top_background=top_genes, group_by=args.channel_col)
    if bg_df.shape[0] >1:
        fig, ax= plt.subplots(figsize=(12,8))
        sns.heatmap(bg_df, ax=ax)
        fig.tight_layout()
        plt.savefig(os.path.join(args.figpath,"heatmap_background_" + args.channel_col + "_rna_top_expressed.png"))
    else:
        fig, ax= plt.subplots(figsize=(12,8))
        plt_df = bg_df.stack().reset_index().rename(columns={'level_1':'feature', 0:'exprs'})
        sns.barplot(data=plt_df, x='feature', y='exprs', ax=ax)
        pnp.plotting.adjust_x_axis(ax)
        fig.tight_layout()
        plt.savefig(os.path.join(args.figpath,"barplot_background_" + args.channel_col + "_rna_top_expressed.png"))


# quantifying the top background features
## Repeat for protein (if it exists)
if 'prot' in mdata.mod.keys():
    # this time we'll just use all the prot vars.
    top_genes = list(mdata_bg['prot'].var_names)
    bg_df = pnp.scmethods.get_mean_background_fraction(mdata_bg['prot'], top_background=top_genes, group_by=args.channel_col)
    if bg_df.shape[0] >1:
        if len(top_genes) > 50:
            # split the plot into two
            split_int = int(np.ceil(137/2))
            fig, ax = plt.subplots(nrows=2,ncols=1,figsize=(24,10), facecolor="white")
            sns.heatmap(bg_df.iloc[:,1:split_int], ax=ax[0])
            sns.heatmap(bg_df.iloc[:,split_int:len(top_genes)], ax=ax[1])
            fig.suptitle("mean exprs (raw counts) of PROT in empty drops")  
            fig.tight_layout()
        else:
            fig, ax = plt.subplots(figsize=(12,10), facecolor="white")
            sns.heatmap(bg_df, ax=ax)
            fig.tight_layout()  
        plt.savefig(os.path.join(args.figpath,"heatmap_background_" + args.channel_col + "_prot_top_expressed.png"))
    else:
        plt_df = bg_df.stack().reset_index().rename(columns={'level_1':'feature', 0:'exprs'})
        if len(top_genes) > 50:
            split_int = int(np.ceil(137/2))
            fig, ax = plt.subplots(nrows=2,ncols=1,figsize=(24,10), facecolor="white")
            sns.heatmap(plt_df.iloc[1:split_int,:], ax=ax[0])
            pnp.plotting.adjust_x_axis(ax[0])
            sns.heatmap(plt_df.iloc[split_int:len(top_genes),:], ax=ax[1])
            pnp.plotting.adjust_x_axis(ax[1])
            fig.suptitle("mean exprs (raw counts) of PROT in empty drops")  
            fig.tight_layout()
        else:
            fig, ax= plt.subplots(figsize=(12,8))
            sns.barplot(data=plt_df, x='feature', y='exprs', ax=ax)
            pnp.plotting.adjust_x_axis(ax)
            fig.tight_layout()
        plt.savefig(os.path.join(args.figpath,"barplot_background_" + args.channel_col + "_prot_top_expressed.png"))


## QC for rna and protein comparing foreground and background

if 'rna' in mdata.mod.keys():
    pnp.pl.scatter_fg_vs_bg(mdata, mdata_bg,x="rna:log1p_n_genes_by_counts", y="rna:log1p_total_counts", facet_row=args.channel_col)
    plt.subplots_adjust(right=0.7)
    plt.savefig(os.path.join(args.figpath,"scatter_bg_fg_rna_nGene_rna_nUMI.png"),transparent=False)
    if 'prot' in mdata.mod.keys():
        if n_samples_rna != n_samples_prot:
            L.info("n_samples are not equal in rna and prot,taking intrsect for mdata and mdata_bg")
            mu.pp.intersect_obs(mdata_bg)
            mu.pp.intersect_obs(mdata)
            pnp.pl.scatter_fg_vs_bg(mdata, mdata_bg,x="prot:log1p_total_counts", y="rna:log1p_n_genes_by_counts", facet_row=args.channel_col)
            plt.subplots_adjust(right=0.7)
            plt.savefig(os.path.join(args.figpath, "scatter_bg_fg_prot_nUMI_rna_nGene.png"),transparent=False)
            pnp.pl.scatter_fg_vs_bg(mdata, mdata_bg,x="rna:log1p_total_counts", y="prot:log1p_total_counts", facet_row=args.channel_col)
            plt.subplots_adjust(right=0.7)
            plt.savefig(os.path.join(args.figpath,"scatter_bg_fg_rna_nUMI_prot_nUMI.png"),transparent=False)
        else:
            L.info("n_samples are equal in rna and prot")
            pnp.pl.scatter_fg_vs_bg(mdata, mdata_bg,x="prot:log1p_total_counts", y="rna:log1p_n_genes_by_counts", facet_row=args.channel_col)
            plt.subplots_adjust(right=0.7)
            plt.savefig(os.path.join(args.figpath, "scatter_bg_fg_prot_nUMI_rna_nGene.png"),transparent=False)
            pnp.pl.scatter_fg_vs_bg(mdata, mdata_bg,x="rna:log1p_total_counts", y="prot:log1p_total_counts", facet_row=args.channel_col)
            plt.subplots_adjust(right=0.7)
            plt.savefig(os.path.join(args.figpath,"scatter_bg_fg_rna_nUMI_prot_nUMI.png"),transparent=False)



plt.clf()


L.info("Done")




import scanpy as sc
import scIB as scib
import glob
import pandas as pd
import argparse
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata
from panpipes.funcs.scmethods import merge_consensus_clust

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--uncorrected_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--batch_corrected_anndata', default='tmp/harmony_scaled.h5ad',
                    help='')
parser.add_argument('--use_muon',
                    default=False,
                    help='') 
parser.add_argument('--integration_method', default='batch',
                    help='')
parser.add_argument('--integration_col', default='batch',
                    help='')
parser.add_argument('--rough_consensus_clustering', default=None,
                    help='must match the barcodes in the anndata objects')
parser.add_argument('--reference_col', default=None,
                    help='')
parser.add_argument('--outfile', default='scib.tsv',
                    help='')
args = parser.parse_args()
args, opt = parser.parse_known_args()
use_muon = check_for_bool(args.use_muon)

L.info("reading data and starting integration pipeline with script: ")
L.info("Running with options: %s", args)

# which metrics to test?
batch_metrics = {
    "isolated_labels_asw_": False,
    "hvg_score_": True,
    "pcr_": True,
    "cell_cycle_": True,
    "ilisi_": True,

}
bio_metrics = {
    "clisi_": True,
    "nmi_": True,
    "ari_": True,
    "isolated_labels_f1_": True,
    "silhouette_": True,
    "graph_conn_": True,
    "kBET_": True,
}
adata = read_anndata(args.uncorrected_anndata, use_muon=use_muon, modality="rna")
# this is just an anndata object
adata_int = read_anndata(args.batch_corrected_anndata, use_muon=False, modality="rna")


# get rough consensus clustering
if args.reference_col is not None:
    if args.rough_consensus_clustering is not None:
        consensus_clust = pd.read_csv(args.rough_consensus_clustering, index_col=0)
        # merge in the consensus clustering
        adata = merge_consensus_clust(adata, consensus_clust)
        adata = adata[adata.obs[args.reference_col].notnull(),:]
        # merge in the consensus clustering
        adata_int = merge_consensus_clust(adata_int, consensus_clust)
        adata_int = adata_int[adata_int.obs[args.reference_col].notnull(),:]
    else:
        # check that reference col is found in the adata
        if args.reference_col not in adata.obs.columns:
            sys.exit("reference column not found in adata object")
    metrics_kwargs = {**batch_metrics, **bio_metrics}
    key_kwargs={"batch_key": args.integration_col, 
                "label_key": args.reference_col}  
else:
    metrics_kwargs = batch_metrics
    key_kwargs={"batch_key": args.integration_col,
            "label_key": "sample_id"}   

if args.integration_method == "harmony":
    x_embed = "X_harmony"
elif args.integration_method == "scanorama":
    x_embed = "X_scanorama"
elif args.integration_method == "scvi":
    x_embed = "X_scVI"
else:
    x_embed = "X_pca"

 
L.info("running scib on")
L.info(metrics_kwargs)
metrics = scib.metrics.metrics(adata=adata,
                        adata_int=adata_int, 
                        cluster_key="leiden",
                         embed=x_embed, 
                         organism="human", **key_kwargs, **metrics_kwargs)

metrics.to_csv(args.outfile)


import scanpy as sc
import glob
import pandas as pd
import argparse
import re
from  muon import MuData
import muon as mu

import warnings
# warnings.filterwarnings("ignore", category=UserWarning)
# warnings.filterwarnings("ignore", category=FutureWarning)

from panpipes.funcs.processing import  concat_mdatas

pd.set_option('display.max_rows', None)
# import os
# os.chdir("../../data/run_qc_general")
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)



parser = argparse.ArgumentParser()
parser.add_argument('--input_files_str',
                    default='',
                    help='')
parser.add_argument('--output_file',
                    default='',
                    help='mdata.h5mu')
parser.add_argument("--submissionfile",
                    default="caf_example",
                    help="this file has all the samples of the experiment. it has at least 3 columns: \
                    sample_id,path and type. \
                    it can contain other anonymous columns which will be renamed using the metadatacols information")
parser.add_argument("--sampleprefix",
                    default="",
                    help="prefix to prepend when saving the metadata file")
parser.add_argument("--metadatacols",
                    default=None,
                    help="names of the metadata to rename the columns of the sample submission file ")
parser.add_argument("--join_type",
                    default="inner",
                    help="whether to concat anndata on the 'inner' (default) or 'outer'")
parser.add_argument('--barcode_mtd_df',
                    default=None,
                    help='csv file contining barcode level metadata')
parser.add_argument('--barcode_mtd_metadatacols',
                    default=None,
                    help='comma separated strings listing the column you want to keep in barcode_mtd_df')            
parser.add_argument('--protein_var_table',
                    default=None,
                    help='')
parser.add_argument('--protein_new_index_col',
                    default=None,
                    help='')        

parser.set_defaults(verbose=True)
args, opt = parser.parse_known_args()
L.info("running with:")
L.info(args)

if len(args.input_files_str) == 0:
    sys.exit("no input files detected")
    
lf = re.split(',', args.input_files_str)

sfile=args.submissionfile

#  lf = glob.glob("./tmp/*raw.h5mu")
# sfile= "CARTC_bonemarrow_samples.tsv"

caf = pd.read_csv(sfile, sep="\t")
# the modality argument is relevent only if use_muon is True.
# mdatas = [read_anndata(i, use_muon, modality="all") for i in lf]
mdatas = [mu.read(i) for i in lf]
# [x.var_names_make_unique() for x in mdatas]


if len(mdatas)==1:
    # if mdatas length is 0 then no concatenation required
    mdata = mdatas[0]
else:
    # @Fabiola I don't understand whats supposed to happen here
    # with the atac, should it just error out to say
    # that atac shouldn't be concatenated?
    temp = mdatas[0]
    if 'atac' in temp.mod.keys():
        mdata=mdatas[0]
        del temp
    elif 'prot' in temp.mod.keys() or 'rna' in temp.mod.keys():
        ## IF RNA and PROT is ok to concatenate ----------
        mdata = concat_mdatas(mdatas,
            batch_key="sample_id", 
            join_type=args.join_type)

# remove to avoid mem issues
del mdatas

### add metmdata from caf
L.debug(mdata.obs.columns)
L.debug(mdata.obs.head())

L.info("add metadata")
# add metmdata to each object
metadatacols = args.metadatacols
# metadatacols="mdata_cols"
# check sample_id is in metadatacols as a minimum requirement
if metadatacols == None :
    pass
elif metadatacols == 'None':
    pass
else:
    metadatacols = metadatacols.split(',')
    if 'sample_id' not in metadatacols:
        metadatacols = ['sample_id'] + metadatacols
    # check all metadatacols are in the caf
    if all([x in caf.columns for x in metadatacols]):
        # if mdata is actually a mudata it will be helpful 
        # (although annoying to store this in rna and protein as 
        # well since the batch columns etc will be here)
        # hopefullt his whole bit will become unnesacry because of 
        # https://github.com/scverse/mudata/issues/2
        for mod in mdata.mod.keys():
            L.debug(mod)
            L.debug(mdata[mod].obs.head())
            assay = mdata[mod]
            # make sure there is no index name because rep might have one as default
            assay.obs.index.name = None
            assay.obs = assay.obs.reset_index().merge(caf[metadatacols], on="sample_id", how="left").set_index('index').copy()
            # remove the index name introduced with the set index.
            assay.obs.index_name = None
            L.debug(assay.obs.head())
        mdata.update_obs()
        # merge metadata
        # obs = mdata.obs
        # obs['cellbarcode'] = obs.index  
        mdata.obs=mdata.obs.reset_index().merge(caf[metadatacols], on="sample_id", how="left").set_index('index')
        mdata.obs.index.name = None
        mdata.obs['cellbarcode'] = mdata.obs.index
        mdata.update_obs()
    else:
        # which column is missing?
        missingcols= " ".join([x for x in metadatacols if x not in caf.columns])
        L.error("Required columns missing form samples file: %s" % missingcols)
        sys.exit("Exiting because required columns not in samples file, check the log file for details")


# if 'rep' in mdata.mod.keys():
#     col_update = mdata['rep'].obs.columns[mdata['rep'].obs.columns.str.contains("count")]
#     mdata['rep'].obs[col_update] = mdata['rep'].obs[col_update].apply(pd.to_numeric)
#     mdata.update()
    
# add in cell level metadata
if args.barcode_mtd_df is not None :
    L.info("Add barcode level metadata")
    # check demult_metadatacols exists and contains antibody
    barcode_metadatacols = args.barcode_mtd_metadatacols.split(',')
    # load the demultiplexing data
    barcode_mtd_df = pd.read_csv(args.barcode_mtd_df, index_col=None)
    if len(barcode_mtd_df['sample_id'].unique().tolist()) > 1:
        barcode_mtd_df.index = barcode_mtd_df[['barcode_id', 'sample_id']].apply(lambda row: "-".join(row.values.astype(str)), axis=1)
    else:
        barcode_mtd_df.index = barcode_mtd_df['barcode_id']
    barcode_mtd_df= barcode_mtd_df.drop(columns=['sample_id'])
    # add for each modality
    for mod in mdata.mod.keys():
        mdata[mod].obs = mdata[mod].obs.merge(barcode_mtd_df, left_index=True, right_index=True)
        mdata.update()
    # add at top level
    mdata.obs = mdata.obs.merge(barcode_mtd_df, left_index=True, right_index=True)


L.debug(mdata.obs.columns)
L.debug(mdata.obs.head())
# update the protein variable to add in extra info like isotype and alternate name for prot
if args.protein_var_table is not None:
    try:
        df = pd.read_csv(args.protein_var_table, sep='\t', index_col=0)
        L.info("merging protein table with var")
        # add_var_mtd(mdata['prot'], df)
        var_df = mdata['prot'].var.merge(df, left_index=True, right_index=True)
        if args.protein_new_index_col is not None:
            L.info("updating prot.var index")
            # update_var_index(mdata['prot'], args.protein_new_index_col)
            var_df =var_df.reset_index().set_index(args.protein_new_index_col)
            var_df = var_df.rename(columns={'index':'orig_id'})
            var_df.index.name = None
        mdata['prot'].var = var_df
        mdata.update_var()
        mdata.update()
        # we might want to split hashing antibodies into a separate modalities
        # we assume this has been inidicated in a "hashing_ab" column in the protein metadata file
        if "hashing_ab" in mdata['prot'].var.columns:
            # create new modality for hashing
            mdata.mod["hashing_ab"]=mdata["prot"][:, mdata["prot"].var["hashing_ab"]]
            # subset old modality to remove hashing
            mdata.mod['prot'] = mdata["prot"][:, ~mdata["prot"].var["hashing_ab"]]
    except FileNotFoundError:
        warnings.warn("protein metadata table not found")
mdata.update()
# tidy up metadata
# move sample_id to the front
# cols = mdata.obs.columns.tolist()
# cols.insert(0, cols.pop(cols.index('sample_id')))
# mdata.obs = mdata.obs.reindex(columns=cols)
L.debug(mdata.obs.dtypes)

L.info("writing to file {}".format(str(args.output_file)))

mdata.write(args.output_file)

L.info("done")




import scanpy as sc
import os
import argparse
import muon as mu
from muon import atac as ac
from panpipes.funcs.processing import check_for_bool


import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument("--input_mudata",
                    default="adata_scaled.h5ad",
                    help="")
parser.add_argument('--modality',
                    default=False,
                    help='')
parser.add_argument("--fig_dir", default="figures/")
# pca options
parser.add_argument("--n_pcs", default=50)
parser.add_argument("--color_by", default="batch")

args, opt = parser.parse_known_args()

sc.settings.autoshow=False
sc.settings.figdir=args.fig_dir 
sc.settings.set_figure_params(dpi_save=300)


mdata = mu.read(args.input_mudata)


if args.color_by is not None:
    qc_vars=args.color_by.split(",")
    qc_vars= [a.strip() for a in qc_vars]
else:
    qc_vars = ["sample_id"]

my_dict = {"rna": 'gex', "prot": 'adt', "atac": 'atac', "rep": 'rep'}
for mod in mdata.mod.keys():
    sc.settings.figdir=os.path.join(args.fig_dir,my_dict[mod])
    pointsize = 120000 / adata.shape[0]
    if "X_pca" in mdata[mod].obsm.keys():
        L.info("Found a PCA in %s, Plotting it" % mod)
        sc.pl.pca(mdata[mod], color=qc_vars, size=pointsize)
        
    if "X_lsi" in mdata[mod].obsm.keys(): 
        L.info("Found a LSI in %s, Plotting it" % mod)
        ac.pl.lsi(mdata[mod],color=qc_vars)
        #https://satijalab.org/signac/reference/DepthCor.html
        #muon.pl.scatter(mdata[mod], x=)

L.info("Done")




'''
scanpy QC script RNA
order of QC:
- RNA
- PROT
- Repertoire
- ATAC
'''
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

import pandas as pd
import os
import argparse
import scanpy as sc
import muon as mu

from panpipes.funcs.io import write_obs


parser = argparse.ArgumentParser()
# required option
parser.add_argument("--sampleprefix",
                    default="",
                    help="prefix to prepend when saving the metadata file")
parser.add_argument("--input_anndata",
                    default="adata_unfilt.h5ad",
                    help="")
parser.add_argument("--outfile",
                    default="adata_unfilt.h5ad",
                    help="")
parser.add_argument("--figdir",
                    default="./figures/",
                    help="path to save the figures to")
parser.add_argument("--figure_suffix",
                    default="_qc-plot.png",
                    help="figures filename suffix to be appended to figures/umap")
parser.add_argument("--scrubletdir",
                    default=None,
                    help="path to save the figures to")
parser.add_argument("--ccgenes",
                    default=None,
                    help="path to file containing cell cycle genes")
parser.add_argument("--customgenesfile",
                    default=None,
                    help="path to file containing list of genes to quantify")
parser.add_argument("--calc_proportions",
                    default="mitochondrial,ribosomal",
                    help="which list of genes to use to calc proportion of mapped reads over total,per cell?")
parser.add_argument("--score_genes",
                    default="MarkersNeutro",
                    help="which list of genes to use to scanpy.tl.score_genes per cell?")

args, opt = parser.parse_known_args()

L.info("Running scanpy rna qc pipeline")

sc.settings.verbosity = 3


L.info("running with args:")
L.info(args)
figdir = args.figdir

if not os.path.exists(figdir):
    os.mkdir(figdir)

sc.settings.figdir = figdir
sc.set_figure_params(scanpy=True, fontsize=14, dpi=300, facecolor='white', figsize=(5,5))


L.info("reading in data")

mdata = mu.read(args.input_anndata)
rna = mdata['rna']


L.info("merge in the scrublet scores")
# load the scrublet scores into the anndata (if they have been run)
if args.scrubletdir is not None:
    scrub_dir = args.scrubletdir
    sample_ids = rna.obs[['sample_id']].drop_duplicates()
    [scrub_dir + '/' + ss + '_scrublet_scores.txt' for ss in sample_ids.sample_id]
    doubletscores = [pd.read_csv(scrub_dir + '/' + ss + '_scrublet_scores.txt', sep="\t", header=0) for ss in sample_ids.sample_id]
    doubletscores = pd.concat(doubletscores, keys=sample_ids.sample_id).reset_index(level="sample_id")
    # rename the barcodes to match up qwith what the rna.obs barcodes are
    if len(sample_ids) > 1:
        doubletscores['barcode'] = doubletscores['barcode'] + '-' + doubletscores['sample_id']
    doubletscores = doubletscores.set_index('barcode').drop('sample_id', axis=1)
    # merge with rna.obs
    rna.obs = rna.obs.merge(doubletscores, how="left", left_index=True, right_index=True)


# Cell-wise QC based on .var lists of genes 

qc_vars = []

if args.customgenesfile is not None:
    if os.path.exists(args.customgenesfile):
        cat_dic = {}
        customgenes = pd.read_csv(args.customgenesfile)
        custom_cat = list(set(customgenes['group'].tolist()))
        for cc in custom_cat:
            cat_dic[cc] = customgenes.loc[customgenes["group"] == cc,"feature"].tolist()
        if args.calc_proportions is not None:
            calc_proportions = args.calc_proportions.split(",")
            calc_proportions = [a.strip() for a in calc_proportions]
        else:
            L.warn("I don't have genes to calc proportions for")
            
        if args.score_genes is not None:
            score_genes = args.score_genes.split(",")
            score_genes = [a.strip() for a in score_genes]
        else:
            L.warn("I don't have genes to calc scores for")    
        
    else:

        sys.exit("You have not provided a list of custom genes to use for QC purposes")

for kk in calc_proportions:
    xname= kk
    gene_list = cat_dic[kk]
    rna.var[xname] = [x in gene_list for x in rna.var_names] # annotate the group of hb genes as 'hb'
    qc_vars.append(xname)

sc.pp.calculate_qc_metrics(rna, qc_vars=qc_vars, percent_top=None,log1p=True, inplace=True)

if args.score_genes is not None:
    for kk in score_genes:
        xname= kk
        gene_list = cat_dic[kk]
        sc.tl.score_genes(rna, gene_list , 
                            ctrl_size=min(len(gene_list), 50), 
                            gene_pool=None, 
                            n_bins=25, 
                            score_name=kk + '_score', 
                            random_state=0, 
                            copy=False, 
                            use_raw=None)


if args.ccgenes is not None:
    ccgenes = pd.read_csv(args.ccgenes, sep='\t')
    sgenes = ccgenes[ccgenes["cc_phase"] == "s"]["gene_name"].tolist()
    g2mgenes = ccgenes[ccgenes["cc_phase"] == "g2m"]["gene_name"].tolist()
    sc.tl.score_genes_cell_cycle(rna, s_genes=sgenes, g2m_genes=g2mgenes)



L.info("calculated scores and metrics")

mdata.update()

L.info("saving anndata and obs in a metadata tsv file")
write_obs(mdata, output_prefix=args.sampleprefix, 
        output_suffix="_cell_metadata.tsv")

mdata.write(args.outfile)

L.info("done")




import pandas as pd
import scanpy as sc
import argparse
import os 
import muon as mu

from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)


# there is a problem with numvba pushong a warning about TBB which is causing the pipeline to crash
# we might need to need to suppress this warning

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--dimred',
                    default='PCA',
                    help='which dimred to expect, relevant for ATAC')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_bbknn.csv',
                    help='')
parser.add_argument('--integration_col', default='batch',
                    help='')
parser.add_argument('--neighbors_within_batch', default='3',
                    help='How many top neighbours to report for each batch; total number of neighbours will be this number times the number of batches.')
parser.add_argument('--neighbors_n_pcs', default='50',
                    help='How many top neighbours to report for each batch; total number of neighbours will be this number times the number of batches.')
parser.add_argument('--modality', default='',
                    help='which mod to run bbknn on')


args, opt = parser.parse_known_args()



L.info("reading data and starting integration pipeline with script: ")
L.info(os.path.basename(__file__))
L.info("Running with options: %s", args)

#adata = read_anndata(args.input_anndata, use_muon=use_muon, modality="rna")
mdata = mu.read(args.input_anndata)
adata = mdata.mod[args.modality] 
L.info("read files")
nnb = int(args.neighbors_within_batch)
# bbknn can't integrate on 2+ variables, so create a fake column with combined information
columns = [x.strip() for x in args.integration_col.split(",")]

if args.modality =="atac":
    if "scaled_counts" in adata.layers.keys():
        pass
    else:
        L.info("to run BBKNN on atac we need to compute PCA. Computing on the flight now")
        sc.pp.scale(adata)
        adata.layers["scaled_counts"] = adata.X.copy()
        sc.tl.pca(adata, n_comps=min(50,adata.var.shape[0]-1), svd_solver='arpack', random_state=0) 

if "X_pca" not in adata.obsm:
    L.info("i need a dimred to start, computing pca with default param")
    n_pcs = 50
    if adata.var.shape[0] < n_pcs:
        L.info("You have less features than number of PCs you intend to calculate")
        n_pcs = adata.var.shape[0] - 1
        L.info("Setting n PCS to %i" % int(n_pcs))    
    sc.pp.scale(adata)
    sc.tl.pca(adata, n_comps=n_pcs, 
                    svd_solver='arpack', 
                    random_state=0) 


L.info("preparing for integration")
if len(columns) > 1:
    L.info("using 2 columns to integrate on more variables")
    # comb_columns = "_".join(columns)
    adata.obs["comb_columns"] = adata.obs[columns].apply(lambda x: '|'.join(x), axis=1)

    # make sure that batch is a categorical
    adata.obs["comb_columns"] = adata.obs["comb_columns"].astype("category")
    # run bbknn
    adata = sc.external.pp.bbknn(adata, batch_key="comb_columns", copy=True, neighbors_within_batch=nnb)
else:
    adata.obs[args.integration_col] = adata.obs[args.integration_col].astype("category")
    # run bbknn
    adata = sc.external.pp.bbknn(adata,
                        batch_key=args.integration_col,
                        copy=True,
                        n_pcs = int(args.neighbors_n_pcs),
                        neighbors_within_batch=nnb)  # calculates the neighbours

L.info("integration run now calculate umap")
sc.tl.umap(adata)
L.info("done umap, saving stuff")
# write out
umap = pd.DataFrame(adata.obsm['X_umap'], adata.obs.index)
umap.to_csv(args.output_csv)



# save full bbknn anndata in tmp, cause need more than just neighbors to work 
outfiletmp = ("tmp/bbknn_scaled_adata_" + args.modality + ".h5ad" )

L.info("saving bbknn corrected adata")
write_anndata(adata, outfiletmp, use_muon=False, modality=args.modality)

L.info("Done")





if args.generate_scanvi:
    if args.adata_reference is not None:
        reference_data = os.path.basename(args.adata_reference)
        mdata=read_anndata(args.adata_reference, use_muon=True, modality="all")
        if reference_data.endswith(".h5mu"):
            if "rna" not in mdata.mod.keys():
                sys.exit("we only support querying using RNA but your mdata doesn't contain rna")
            else:
                adata_ref = mdata["rna"].copy()
        else:
            adata_ref = mdata.copy()

        adata_ref.obs.loc[:, 'is_reference'] = 'Reference'
        adata_query.obs.loc[:, 'is_reference'] = 'Query'
        adata_query = adata_query[:, adata_ref.var_names].copy()
     
        #just to be clear, what i don't understand is why i can't split a scvi training from a scanvi update by saving the model and realoading afterwards. 
        #if i want to create a scanvi model from a scvi saved model there is no way to do it unless i retrain from scratch
        L.info(""" Generating scanvi reference from initialized SCVI, 
                    we'll train a scvi model first and
                    we expect the cell type labels into a cell_type column
                    """)   
        scvi.model.SCVI.setup_anndata(adata_ref, batch_key="tech", layer="counts")
        
        arches_params = dict(
            use_layer_norm="both",
            use_batch_norm="none",
            encode_covariates=True,
            dropout_rate=0.2,
            n_layers=2,
        )

        vae_ref = scvi.model.SCVI(
            adata_ref,
            **arches_params
        )
        vae_ref.train()    
        os.makedirs("models")
        dir_path_scan=os.path.join("models", "scvi_model")
        vae_ref.save(dir_path_scan, anndata=False)

        vae_ref_scan = scvi.model.SCANVI.from_scvi_model(
            vae_ref, #this is scvi reference model
            unlabeled_category="Unknown", #how to call the cells with no label in the "cell_type" column
            labels_key="cell_type")
        vae_ref_scan.train(max_epochs=20, n_samples_per_label=100) #to do add training args options
        
        dir_path_scan=os.path.join("models", "scanvi_model")
        vae_ref_scan.save(dir_path_scan, anndata=False)
        
        vae_q = scvi.model.SCANVI.load_query_data(
        adata_query,
        dir_path_scan
        )

        vae_q._unlabeled_indices = np.arange(adata_query.n_obs)
        vae_q._labeled_indices = []
        latent_choice= "X_scANVI"
        vae_q.train(**train_kwargs)
        adata_query.obsm["X_scANVI"] = vae_q.get_latent_representation()
        adata_query.obs["predictions"] = vae_q.predict()
        adata_full = adata_query.concatenate(adata_ref)
    else:
        sys.exit("To generate a scanvi reference i need a reference dataset to start from")
        

L.info("Done")




'''
run scrublet on single channel
expects sample id and path to input data
'''
import scrublet as scr
import pandas as pd
import os
import argparse
import muon as mu

import sys
import logging





import matplotlib
matplotlib.use('Agg')
matplotlib.pyplot.ioff()

L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--sample_id",
                    default="sampleID",
                    help="name of the sample, usually just the name of the cellranger filtered folder")
parser.add_argument("--inputpath",
                    default="/path/to/anndata_object",
                    help="path to the single channel worth of cells in anndata format")
parser.add_argument("--outdir",
                    default="/gpfs3/well/combat/projects/preprocess/citeseq_final/FC_annotation/dotplot_minimal/",
                    help="string, b or t cells?")
parser.add_argument("--expected_doublet_rate",
                    default=0.06,
                    help="the expected fraction of transcriptomes that are doublets, typically 0.05-0.1. Results are not particularly sensitive to this parameter")
parser.add_argument("--sim_doublet_ratio",
                    default=2,
                    help="the number of doublets to simulate, relative to the number of observed transcriptomes. Setting too high is computationally expensive. Min tested 0.5")
parser.add_argument("--n_neighbors",
                    default=20,
                    help="Number of neighbors used to construct the KNN classifier of observed transcriptomes and simulated doublets. The default value of round(0.5*sqrt(n_cells)) generally works well")
parser.add_argument("--min_counts",
                    default=2,
                    help="Used for gene filtering prior to PCA. Genes expressed at fewer than `min_counts` in fewer than `min_cells` (see below) are excluded")
parser.add_argument("--min_cells",
                    default=3,
                    help="Used for gene filtering prior to PCA. Genes expressed at fewer than `min_counts` (see above) in fewer than `min_cells` are excluded.")
parser.add_argument("--min_gene_variability_pctl",
                    default=85,
                    help="Used for gene filtering prior to PCA. Keep the most highly variable genes (in the top min_gene_variability_pctl percentile), as measured by the v-statistic [Klein et al., Cell 2015]")
parser.add_argument("--n_prin_comps",
                    default=30,
                    help="Number of principal components used to embed the transcriptomes prior to k-nearest-neighbor graph construction")
parser.add_argument("--use_thr",
                    default=True,
                    help="use a user defined thr to define min doublet score to split true from false doublets? if false just use what the software produces")
parser.add_argument("--call_doublets_thr",
                    default=0.25,
                    help="if use_thr is True, this thr will be used to define doublets")
args, opt = parser.parse_known_args()

# loading data
adata = mu.read(args.inputpath + "/rna")


    
counts_matrix = adata.X
L.info('Counts matrix shape: {} rows, {} columns'.format(counts_matrix.shape[0], counts_matrix.shape[1]))
L.info('Number of genes in gene list: {}'.format(len(adata.var_names)))

L.info("now initializing the scrublet object with expected_doublet_rate:\n")
L.info(args.expected_doublet_rate)

scrub = scr.Scrublet(counts_matrix, expected_doublet_rate=float(args.expected_doublet_rate))
L.info("predicting doublets with params: \nmincells: %s \nmingenes: %s \nmin_gene_variabilty_pctl: %s\nn_prin_comps: %s\n" % (args.min_counts, args.min_cells, args.min_gene_variability_pctl, args.n_prin_comps))
L.info("predicting using:\n")

doublet_scores, predicted_doublets = scrub.scrub_doublets(min_counts=int(args.min_counts),
                                                          min_cells=int(args.min_cells),
                                                          min_gene_variability_pctl=float(args.min_gene_variability_pctl),
                                                          n_prin_comps=int(args.n_prin_comps))

# cgat pipelines will probably parse a string
if args.use_thr == "True":
    use_thr = True
else:
    use_thr = False



if use_thr:
    L.info("using provided or default threshold  %s to call doublets, instead of predicted" % args.call_doublets_thr)
    predicted_doublets = scrub.call_doublets(threshold=float(args.call_doublets_thr))
else:

    L.info("attempting to use predicted threshold for doublet prediction")
    try:
        use_threshold=round(scrub.threshold_, 2)
        L.info("using predicted threshold %s" % use_threshold)
    except AttributeError:
        use_threshold=float(args.call_doublets_thr)
        L.info("scrublet couldn't predict a threshold falling back on provided or default threshold %s" % use_threshold)

    predicted_doublets = scrub.call_doublets(threshold=float(use_threshold))
L.info("saving plots to outdir")

fig = scrub.plot_histogram()
fig[0].savefig(args.outdir + '/' + args.sample_id + '_' + "doubletScore_histogram.png",
               bbox_inches='tight', dpi=120)

L.info("cellnames and doublet scores and prediction")

data = pd.DataFrame({'doublet_scores': doublet_scores,
                     'predicted_doublets': predicted_doublets})

data['barcode'] = adata.obs_names
data.to_csv(os.path.join(args.outdir + "/" + args.sample_id + "_scrublet_scores.txt"),
            sep="\t", index=False)
L.info('done')






# check for threads number
import multiprocessing 
threads_available = multiprocessing.cpu_count()

# import numpy as np
import pandas as pd
import scanpy as sc
import os
import argparse
from muon import MuData
from panpipes.funcs.processing import check_for_bool
from panpipes.funcs.io import read_anndata, write_anndata
from panpipes.funcs.scmethods import run_neighbors_method_choice
import muon as mu
import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse arguments
parser = argparse.ArgumentParser()

parser.add_argument('--input_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--modality',
                    default='rna',
                    help='')
parser.add_argument('--dimred',
                    default='PCA',
                    help='which dimred to expect, relevant for ATAC')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_none.csv',
                    help='')
parser.add_argument('--n_threads', default=1,
                    help="num threads to use for neighbor computations")
parser.add_argument('--integration_col', default='batch')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs")
parser.add_argument('--neighbors_method',
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k',
                    help="neighbors k")
parser.add_argument('--neighbors_metric',
                    help="neighbor metric, e.g. euclidean or cosine")

args, opt = parser.parse_known_args()

L.info("reading data and starting integration pipeline with script: ")
L.info(os.path.basename(__file__))


# Scanorama is designed to be used in scRNA-seq pipelines downstream of noise-reduction methods,
# including those for imputation and highly-variable gene filtering.

# adata = sc.read(args.input_anndata)
#adata = read_anndata(args.input_anndata, use_muon=True, modality=args.modality)
adata = mu.read(args.input_anndata +"/" + args.modality)
# adata = mdata.mod[args.modality] 



columns = [x.strip() for x in args.integration_col.split(",")]
if len(columns)>1: 
    comb_columns = "|".join(columns)
    adata.obs[comb_columns] = adata.obs[columns].apply(lambda x: '|'.join(x), axis=1)
    columns += [comb_columns]

# write out batch
# adata.obs[columns].to_csv(os.path.join(os.path.dirname(args.output_csv), 'batch_'+ args.modality +'_mtd.csv'))
if args.dimred == "PCA":
    dimred = "X_pca"
elif args.dimred == "LSI":
    dimred = "X_lsi"


# run neighbours and umap without batch correction
pc_kwargs = {}
if int(args.neighbors_n_pcs) > 0:
    pc_kwargs['use_rep'] = dimred
    pc_kwargs['n_pcs'] = int(args.neighbors_n_pcs)
else:
    # need to push scanpy to use .X if use_rep is None.
    pc_kwargs['use_rep'] = None
    pc_kwargs['n_pcs'] = int(args.neighbors_n_pcs)


if dimred not in adata.obsm:
    L.info("i need a dimred to start, computing pca with default param")
    n_pcs = 50
    if adata.var.shape[0] < n_pcs:
        L.info("You have less features than number of PCs you intend to calculate")
        n_pcs = adata.var.shape[0] - 1
        L.info("Setting n PCS to %i" % int(n_pcs))    
    sc.pp.scale(adata)
    sc.tl.pca(adata, n_comps=n_pcs, 
                    svd_solver='arpack', 
                    random_state=0) 
    pc_kwargs['use_rep'] = "X_pca"
    pc_kwargs['n_pcs'] = n_pcs



run_neighbors_method_choice(adata, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    metric=args.neighbors_metric, 
    nthreads=max([threads_available, 6]), **pc_kwargs)

L.info("done n_neighbours, saving stuff")

sc.tl.umap(adata)

L.info("done umap, saving stuff")
#write out
umap = pd.DataFrame(adata.obsm['X_umap'], adata.obs.index)
umap.to_csv(args.output_csv)

L.info("done")


outfiletmp = ("tmp/no_correction_scaled_adata_" + args.modality + ".h5ad" )

L.info("saving harmony corrected adata")
write_anndata(adata, outfiletmp, use_muon=False, modality=args.modality)

L.info("done")




"""
# reads in an anndata object with neighbours already computed, runs scanpy umap
# save umap coordinates (for plotting elsewhere
# CRG 2020-06-12
"""
import scanpy as sc
import numpy as np
import pandas as pd
import argparse
import muon as mu
from anndata import AnnData

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

parser = argparse.ArgumentParser()
parser.add_argument("--infile",
                    default="mdata.h5mu",
                    help="file name, format: .h5mu")
parser.add_argument('--modality',
                    default=None,
                    help='')
parser.add_argument("--outfile", 
                    default="umap.txt.gz", 
                    help="file name, format: .h5ad")
parser.add_argument("--min_dist", 
                    default=0.1, 
                    help="no. neighbours parameters for sc.pp.neighbors()")
parser.add_argument("--neighbors_key", 
                    default="neighbors", help="algortihm choice from louvain and leiden")

args, opt = parser.parse_known_args()
L.info(args)

# read data
mdata = mu.read(args.infile)
if type(mdata) is AnnData:
    adata = mdata
elif args.modality is not None:
    adata = mdata[args.modality]
else:
    adata = mdata
    

# set seed
# seed = int(200612)

uns_key=args.neighbors_key
# check sc.pp.neihgbours has been run
if uns_key not in adata.uns.keys():
    # sys.exit("Error: sc.pp.neighbours has not been run on this object")
    L.warning("running neighbors with default parameters since no neighbors graph found in this data object")
    sc.pp.neighbors(adata)
    uns_key="neighbors"

# code to run multiple umaps in one script, currently not used
# manipulate min dist to list of floats
# if type(args.min_dist) is str:
#     min_dist = [float(md) for md in args.min_dist.split(',')]
# elif type(args.min_dist) is float:
#     min_dist = [args.min_dist]


# what parameters?
if uns_key =="wnn":
    mu.tl.umap(adata, min_dist=float(args.min_dist), neighbors_key=uns_key)
else:
    sc.tl.umap(adata, min_dist=float(args.min_dist), neighbors_key=uns_key)

# extract umap coordinates for plotting (in R??)
umap_coords = pd.DataFrame(adata.obsm['X_umap'])

# add in the rownames 
umap_coords.index = adata.obs_names

# save coordinates to file
# (note this saves values values up to 6 significant figures, because why save 20 for a plot
umap_coords.to_csv(args.outfile, sep = '\t')



import multiprocessing 


import scanpy as sc
import pandas as pd
import matplotlib.pyplot as plt
import scvi
import argparse
import os
import muon as mu

import panpipes.funcs as pp
from panpipes.funcs.io import read_anndata, read_yaml
from panpipes.funcs.scmethods import run_neighbors_method_choice, X_is_raw

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# load arguments 
parser = argparse.ArgumentParser()
parser.add_argument('--scaled_anndata',
                    default='adata_scaled.h5ad',
                    help='')
parser.add_argument('--raw_anndata',
                    default=None,
                    help='')
parser.add_argument('--output_csv', default='batch_correction/umap_bc_scvi.csv',
                    help='')
parser.add_argument('--integration_col', default='batch',
                    help='')
parser.add_argument('--figdir', default='./figures',
                    help='')
parser.add_argument('--neighbors_n_pcs',
                    help="n_pcs")
parser.add_argument('--neighbors_method',
                    help="neighbours method, scanpy or hnsw")
parser.add_argument('--neighbors_k',
                    help="neighbors k")
parser.add_argument('--neighbors_metric',
                    help="neighbor metric, e.g. euclidean or cosine")


args, opt = parser.parse_known_args()
L.info(args)

threads_available = multiprocessing.cpu_count()

# load parameters
params = read_yaml("pipeline.yml")
L.info("Running scvi script")
L.info("threads available: %s", threads_available)

# scanpy settings
sc.set_figure_params(facecolor="white")
sc.settings.autoshow = False
sc.settings.figdir = args.figdir


# test_script=params['rna']['scvi']['testrun']
test_script=False
if test_script:
    L.info("this is a test run")
    params['rna']['scvi']['training_args']['max_epochs'] = 10


# load an process the scaled snRNAseq dataset --------------------------------------------------


L.info("loading in anndata object ")


mdata = mu.read(args.scaled_anndata)
if type(mdata) is mu.MuData:
    rna = mdata['rna']
    L.info(rna)
elif type(mdata) is sc.AnnData:
    rna = mdata
    L.info(rna)
else:
    L.error("unknown file type")
    sys.exit("file not MuData or Anndata")

# this is a copy so we can subset vars and obs without changing the original object


# in case of more than 1 variable, create a fake column with combined information
columns = [x.strip() for x in args.integration_col.split(",")]
if len(columns) > 1:
    L.info("using 2 columns to integrate on more variables")
    # bc_batch = "_".join(columns)
    rna.obs["bc_batch"] = rna.obs[columns].apply(lambda x: '|'.join(x), axis=1)

    # make sure that batch is a categorical
    rna.obs["bc_batch"] = rna.obs["bc_batch"].astype("category")
else:
    rna.obs['bc_batch'] = rna.obs[args.integration_col]


# add in raw counts as a layer 
# add in test to see if raw layer exists alread
if "raw_counts" in rna.layers.keys():
    L.info("raw counts found in layer")
elif X_is_raw(rna):
    rna.layers["raw_counts"] = rna.X.copy()
else:
    L.info("merge in raw counts")
    sc_raw = read_anndata(args.raw_anndata, use_muon=True, modality="rna")
    #filter by barcodes in the scaled object
    sc_raw = sc_raw[sc_raw.obs_names.isin(rna.obs_names),: ]
    rna.layers["raw_counts"] = sc_raw.X.copy()


# filter out mitochondria
if params['rna']['scvi']['exclude_mt_genes']:
    rna = rna[:, ~rna.var[params['rna']['scvi']['mt_column']]]

# filter by Hvgs
rna = rna[:, rna.var.highly_variable]


if test_script:
    nn=1000
    L.info("this is a test so downsampling the dataset to %i cells" % nn)
    sc.pp.subsample(rna, n_obs=nn)

rna = rna.copy()

# scvi.model.data.setup_anndata(
scvi.model.SCVI.setup_anndata(
    rna,
    layer="raw_counts",
    batch_key='bc_batch'
)


scvi_model_args =  {k: v for k, v in params['rna']['scvi']['model_args'].items() if v is not None}
print(scvi_model_args)
scvi_training_args =  {k: v for k, v in params['rna']['scvi']['training_args'].items() if v is not None}
print(scvi_training_args)
scvi_training_plan =  {k: v for k, v in params['rna']['scvi']['training_plan'].items() if v is not None}
print(scvi_training_plan)



vae = scvi.model.SCVI(rna, **scvi_model_args) 

vae.train(**scvi_training_args, plan_kwargs=scvi_training_plan) 

vae.save(os.path.join("batch_correction", "scvi_model"), 
                  anndata=False)

# no early stopping?
vae.history["elbo_train"].plot()
plt.savefig(os.path.join(args.figdir, "scvi_elbo_train.png"))

fig, axs = plt.subplots(nrows=3, ncols=4, figsize=(16,8))
axs = axs.ravel()
for i, kk in enumerate(vae.history.keys()):
    vae.history[kk].plot(ax=axs[i])
    
fig.tight_layout()
plt.savefig(os.path.join(args.figdir, "scvi_metrics.png"))

# vae.history['elbo_train']['elbo_train'].to_list()

latent = vae.get_latent_representation()

rna.obsm["X_scVI"] = latent
rna.obs['bc_batch'] = rna.obs['bc_batch']

sc.pl.embedding(rna, "X_scVI", color="bc_batch", save="_batch.png")

plot_df = pd.DataFrame(latent[:, 0:2])
plot_df['bc_batch'] = rna.obs['bc_batch'].tolist()

if int(args.neighbors_n_pcs) > rna.obsm['X_scVI'].shape[1]:
    L.warn(f"N PCs is larger than X_scVi dimensions, reducing n PCs to  {rna.obsm['X_scVI'].shape[1] -1}")

n_pcs= min(int(args.neighbors_n_pcs), rna.obsm['X_scVI'].shape[1]-1)

print(rna.obsm['X_scVI'].shape)
# use scVI latent space for UMAP generation
# sc.pp.neighbors(rna,n_neighbors=int(args.n_neighbors), use_rep="X_scVI")
run_neighbors_method_choice(rna, 
    method=args.neighbors_method, 
    n_neighbors=int(args.neighbors_k), 
    n_pcs=n_pcs, 
    metric=args.neighbors_metric, 
    use_rep='X_scVI',
    nthreads=max([threads_available, 6]))

sc.tl.umap(rna)

sc.pl.umap(
    rna,
    color=["bc_batch",],
    frameon=False, save="_scvi_batch.png"
)

umap = pd.DataFrame(rna.obsm['X_umap'], rna.obs.index)
umap.to_csv(args.output_csv)

# save anndata to be used by other scvi tools applications
rna.write(os.path.join("tmp", "scvi_scaled_adata_rna.h5ad"))

L.info("Done")




"""
Fetching data for plotting
CRG 2020-06-19
"""

import scanpy as sc
from anndata import AnnData
import muon as mu
import pandas as pd
import argparse

sc.settings.autoshow = False

import matplotlib
matplotlib.use('agg')

import sys
import logging
L = logging.getLogger()
L.setLevel(logging.INFO)
log_handler = logging.StreamHandler(sys.stdout)
formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s')
log_handler.setFormatter(formatter)
L.addHandler(log_handler)

# parse argumetns
parser = argparse.ArgumentParser()
parser.add_argument("--infile",
                    default="mdata.h5mu, full object, not filtered by HVG",
                    help="file name, format: .h5mu")
parser.add_argument('--modality',
                    default=None,
                    help='')
parser.add_argument("--layer",
                    default="logged_counts",
                    help="layer containing data to be plotted, default=X")
parser.add_argument("--group_col",
                    default="cluster",
                    help="column in obs containing groups for plots,")
parser.add_argument("--marker_file",
                    default="",
                    help="markers from sc.rank_genes_group() as txt(.gz)")
parser.add_argument("--figure_prefix", default="figures/markers_",
                    help="figures path")
parser.add_argument("--n", default=None,
                    help="number of genes to plot per cluster, default=None will plot all genes in your marker_file")



args, opt = parser.parse_known_args()
L.info(args)

sc.settings.figdir = args.figure_prefix

# script

def calc_dendrogram(adata, group_col):
    if len(adata.obs[group_col].cat.categories) > 5:
        incl_dendrogram = True
        if "X_pca" not in adata.obsm.keys():
            sc.pp.pca(adata)
        L.info("calculating dendrogram")
        try:
            sc.tl.dendrogram(adata, groupby=group_col, use_rep="X_pca")
        except ValueError:
            L.info("cannot calculate dendrogram")
            incl_dendrogram = False
    else:
        incl_dendrogram = False
    return incl_dendrogram


def do_plots(adata, mod, group_col, mf, n=10, layer=None):
    L.debug("check layers")
    # get markers for plotting
    mf = mf[mf['avg_logFC'] > 0]
    df = mf.groupby(group_col).apply(lambda x: x.nlargest(n, ['scores'])).reset_index(drop=True)
    marker_list={str(k): list(v) for k,v in df.groupby(group_col)["gene"]}
    # add cluseter col to obs
    # check whether a dendrogram is computed/
    incl_dendrogram = calc_dendrogram(adata, group_col)
    L.info("start plotting")
    sc.pl.stacked_violin(adata,
                marker_list,
                groupby=group_col,
                layer=layer,
                save = '_top_markers'+ mod +'.png',
                dendrogram=incl_dendrogram,
                # figsize=(24, 5)
                )
    sc.pl.matrixplot(adata,
                    marker_list,
                    groupby=group_col,
                    save=  '_top_markers'+ mod +'.png',
                    dendrogram=incl_dendrogram,
                    figsize=(24, 5))
    sc.pl.dotplot(adata,
                marker_list,
                groupby=group_col,
                save=  '_top_markers'+ mod +'.png',
                dendrogram=incl_dendrogram,
                figsize=(24, 5))
    sc.pl.heatmap(adata,
                marker_list,
                groupby=group_col,
                save=  '_top_markers'+ mod +'.png',
                dendrogram=incl_dendrogram,
                # figsize=(24, 5)
                )


# read data
mdata = mu.read(args.infile)

if type(mdata) is AnnData:
    adata = mdata
    # main function only does rank_gene_groups on X, so 
elif type(mdata) is mu.MuData and args.modality is not None:
    adata = mdata[args.modality]    
else:
    sys.exit('if inputting a mudata object, you need to specify a modality')
    

L.info("load marker file")
mf = pd.read_csv(args.marker_file, sep='\t' )
mf[args.group_col] = mf['cluster'].astype('category')

# get layer
adata.obs[args.group_col] = mdata.obs[args.group_col]
do_plots(adata,
         mod=args.modality, 
         group_col=args.group_col,
         mf=mf,
         layer=args.layer, 
         n=int(args.n))

L.info("Done")

